import torch

from MLTT.cache import conditional_lru_cache

EPSILON = 1e-6
CACHE_SIZE = 128

@conditional_lru_cache(maxsize=CACHE_SIZE)
def change(prices, lag: int = 1) -> torch.Tensor:
    """
    Computes absolute change in prices.

    Args:
        - `prices` (torch.Tensor): A tensor of shape `(batch_size, n_tradable)`.
        - `lag` (int): The number of periods to lag the prices by.
    Returns:
        torch.Tensor: Absolute change in prices. Shape: `(batch_size, n_tradable)`.
    """
    prices = torch.as_tensor(prices)

    pad = prices[0].detach().repeat(lag, 1)
    shifted_prices = torch.cat((pad, prices), dim=0)

    return shifted_prices[lag:] - shifted_prices[:-lag]

def apply_columnwise(s: torch.Tensor, func: callable) -> torch.Tensor:
    """
    Apply a function column-wise to a tensor.

    Result will be cached.

    Args:
        s (torch.Tensor): The input tensor.
        func (callable): The function to apply column-wise.
    """
    if s.dim() == 1:
        return func(s)

    res = torch.zeros_like(s)

    for j in range(s.shape[1]):
        res[:, j] = func(s[:, j])

    return res

@conditional_lru_cache(maxsize=CACHE_SIZE)
def to_weights_matrix(x: torch.Tensor) -> torch.Tensor:
    """
    Normalize a matrix to have each row sum to one when considering absolute values
    
    Parameters
    ----------
    x : torch.Tensor
        Input matrix
        
    Returns
    -------
    torch.Tensor
        Normalized matrix where each row sums to 1
    """
    # Convert to tensor if needed
    x = torch.as_tensor(x)
    
    # If x is 1D, just use to_weights directly
    if x.dim() == 1:
        return to_weights(x)
    
    # Sum absolute values along dim 1 (columns) for each row
    row_abs_sum = torch.sum(torch.abs(x), dim=1, keepdim=True)
    
    # Avoid division by zero
    row_abs_sum = torch.where(row_abs_sum == 0, torch.ones_like(row_abs_sum), row_abs_sum)
    
    # Normalize each row
    return x / row_abs_sum

@conditional_lru_cache(maxsize=CACHE_SIZE)
def to_weights(x):
    x = torch.as_tensor(x)
    return x / torch.sum(torch.abs(x))

@conditional_lru_cache(maxsize=CACHE_SIZE)
def alpha_beta(series: torch.Tensor, benchmark: torch.Tensor) -> tuple[float, float]:
    """
    Calculate alpha and beta for a series against a benchmark
    
    Parameters
    ----------
    series : torch.Tensor
        Returns series to evaluate
    benchmark : torch.Tensor
        Benchmark returns series
    
    Returns
    -------
    tuple[float, float]
        Alpha and beta values
    """
    # Convert inputs to tensors
    series = torch.as_tensor(series, dtype=torch.float32)
    benchmark = torch.as_tensor(benchmark, dtype=torch.float32)
    
    # Calculate covariance matrix
    stacked = torch.stack([benchmark, series])
    cov_matrix = torch.cov(stacked)
    
    # Extract values
    var_benchmark = cov_matrix[0, 0]
    cov_series_benchmark = cov_matrix[0, 1]
    
    # Calculate beta
    beta = cov_series_benchmark / var_benchmark
    
    # Calculate alpha (mean of series - beta * mean of benchmark)
    alpha = torch.mean(series) - beta * torch.mean(benchmark)
    
    return alpha.item(), beta.item()

@conditional_lru_cache(maxsize=CACHE_SIZE)
def mad_normalize(x: torch.Tensor) -> torch.Tensor:
    """
    Normalize data using Median Absolute Deviation (MAD) method
    
    x_norm = (x - median(x)) / MAD
    where MAD = median(|x - median(x)|)
    
    Parameters
    ----------
    x : torch.Tensor
        Input data
    
    Returns
    -------
    torch.Tensor
        Normalized data
    """
    # Convert to tensor if needed
    x = torch.as_tensor(x)
    
    # Calculate median
    median = torch.median(x)
    
    # Calculate MAD
    mad = torch.median(torch.abs(x - median))
    
    # Handle case where MAD is zero to avoid division by zero
    if mad == 0:
        return torch.zeros_like(x)
    
    # Normalize
    return (x - median) / mad

def ensure_tensor(*arrays, device=None) -> torch.Tensor | tuple[torch.Tensor, ...]:
    """
    Convert arrays to PyTorch tensors with consistent float32 dtype
    
    Parameters
    ----------
    arrays : array-like
        Arrays to convert
    device : torch.device, optional
        Device to place tensors on
        
    Returns
    -------
    torch.Tensor or tuple of torch.Tensor
        Converted tensor(s) with float32 dtype
    """
    tensors = []
    
    for array in arrays:
        if isinstance(array, torch.Tensor):
            tensor = array.float()  # Ensure float32 dtype
        else:
            tensor = torch.as_tensor(array, dtype=torch.float32)  # Convert to float32
            
        if device is not None:
            tensor = tensor.to(device)
            
        tensors.append(tensor)
        
    return tensors[0] if len(tensors) == 1 else tuple(tensors)

@conditional_lru_cache(maxsize=CACHE_SIZE)
def univariate_barrier(x: torch.Tensor, 
                      upper_bound: torch.Tensor | float, 
                      lower_bound: torch.Tensor | float,
                      neutral_level: torch.Tensor | float | None = 0) -> torch.Tensor:
    """
    Generates trading signals based on dynamic upper and lower barriers.
    
    Takes a time series and generates trading signals when the values cross
    specified upper and lower bounds. Implements a mean reversion strategy:
    - Generates -1 (short) when value exceeds upper bound
    - Generates +1 (long) when value falls below lower bound 
    - Maintains previous position when value is between bounds
    - If neutral_level is provided:
        Exits position (0) when value crosses back through neutral level
    - If neutral_level is None:
        Exits position only when opposite signal is generated
        (e.g. short position is closed only when price falls below lower bound)
    
    Parameters
    ----------
    x : torch.Tensor
        Input time series values
    upper_bound : torch.Tensor or float
        Upper barrier threshold, can be dynamic (tensor) or static (float)
    lower_bound : torch.Tensor or float
        Lower barrier threshold, can be dynamic (tensor) or static (float)
    neutral_level : torch.Tensor or float or None, default 0
        Level at which positions are exited, can be dynamic or static.
        If None, positions are held until opposite signal is generated.
        
    Returns
    -------
    torch.Tensor
        Tensor of -1, 0, +1 trading signals
    """
    # Convert inputs to tensors if needed
    x = torch.as_tensor(x)
    
    # Ensure x is at least 1D
    if x.dim() == 0:
        x = x.unsqueeze(0)
    
    signal = torch.zeros_like(x)
    prev_signal = 0
    
    # Convert scalar bounds to tensors
    if isinstance(upper_bound, (int, float)):
        upper_bound = torch.full_like(x, upper_bound)
    # Convert bounds to match time dimension if needed
    elif isinstance(upper_bound, torch.Tensor) and upper_bound.shape[0] == 1:
        upper_bound = upper_bound.expand_as(x)
        
    if isinstance(lower_bound, (int, float)):
        lower_bound = torch.full_like(x, lower_bound)
    elif isinstance(lower_bound, torch.Tensor) and lower_bound.shape[0] == 1:
        lower_bound = lower_bound.expand_as(x)
        
    if neutral_level is not None:
        if isinstance(neutral_level, (int, float)):
            neutral_level = torch.full_like(x, neutral_level)
        elif isinstance(neutral_level, torch.Tensor) and neutral_level.shape[0] == 1:
            neutral_level = neutral_level.expand_as(x)

    for i in range(x.shape[0]):
        if x[i] > upper_bound[i]:
            signal[i] = -1
        elif x[i] < lower_bound[i]:
            signal[i] = 1
        else:
            signal[i] = prev_signal

        # Exit positions based on neutral level if provided
        if neutral_level is not None:
            if signal[i] == -1 and x[i] <= neutral_level[i]:
                signal[i] = 0
            elif signal[i] == 1 and x[i] >= neutral_level[i]:
                signal[i] = 0
        # Otherwise exit only on opposite signals
        else:
            if signal[i] == -1 and x[i] < lower_bound[i]:
                signal[i] = 1  # Direct switch to long
            elif signal[i] == 1 and x[i] > upper_bound[i]:
                signal[i] = -1  # Direct switch to short

        prev_signal = signal[i]

    return signal

@conditional_lru_cache(maxsize=CACHE_SIZE)
def multivariate_barrier(x: torch.Tensor, 
                      upper_bound: torch.Tensor | float, 
                      lower_bound: torch.Tensor | float,
                      volatility: torch.Tensor | float = 1,
                      neutral_level: torch.Tensor | float | None = 0) -> torch.Tensor:
    """
    Performs univariate barrier calculations across multiple columns/series.
    
    Parameters
    ----------
    x : torch.Tensor
        2D input tensor with shape (time, features)
    upper_bound : torch.Tensor or float
        Upper barrier threshold(s)
    lower_bound : torch.Tensor or float
        Lower barrier threshold(s)
    volatility : torch.Tensor or float, default 1
        Volatility scaling for each series
    neutral_level : torch.Tensor or float or None, default 0
        Levels at which positions are exited
        
    Returns
    -------
    torch.Tensor
        2D tensor of -1, 0, +1 trading signals with shape (time, features)
    """
    # Convert inputs to tensors
    x = torch.as_tensor(x)
    
    # Handle single dimension input
    if x.dim() == 1:
        return univariate_barrier(x, upper_bound, lower_bound, neutral_level)
    
    # Create output tensor
    signal = torch.zeros_like(x)
    
    # Handle scalar bounds for all time steps
    if isinstance(upper_bound, (int, float)):
        upper_bound = torch.full_like(x, upper_bound)
    elif upper_bound.dim() == 1 and upper_bound.shape[0] == 1:
        upper_bound = upper_bound.expand_as(x)
    
    if isinstance(lower_bound, (int, float)):
        lower_bound = torch.full_like(x, lower_bound)
    elif lower_bound.dim() == 1 and lower_bound.shape[0] == 1:
        lower_bound = lower_bound.expand_as(x)
    
    if isinstance(volatility, (int, float)):
        volatility = torch.full((1, x.shape[1]), volatility)
    elif volatility.dim() == 1:
        volatility = volatility.unsqueeze(0)
    
    if neutral_level is not None:
        if isinstance(neutral_level, (int, float)):
            neutral_level = torch.full((1, x.shape[1]), neutral_level)
        elif neutral_level.dim() == 1:
            neutral_level = neutral_level.unsqueeze(0)
    
    # Process each column separately
    for j in range(x.shape[1]):
        # Extract bounds for this column, adjusted by volatility
        col_upper = upper_bound[:, j] * volatility[:, j]
        col_lower = lower_bound[:, j] * volatility[:, j]
        
        # Handle neutral level
        col_neutral = None
        if neutral_level is not None:
            col_neutral = neutral_level[:, j]
            
        # Apply univariate barrier to this column
        signal[:, j] = univariate_barrier(
            x[:, j], col_upper, col_lower, col_neutral
        )
        
    return signal


@conditional_lru_cache(maxsize=CACHE_SIZE)
def calculate_betas(asset_returns: torch.Tensor, 
                   benchmark_returns: torch.Tensor,
                   beta_measurement_period: int = None) -> torch.Tensor:
    """
    Calculate beta coefficients for assets relative to benchmark.
    
    Args:
        asset_returns: Tensor of asset returns with shape (time_steps, num_assets)
        benchmark_returns: Tensor of benchmark returns with shape (time_steps, 1)
        beta_measurement_period: Number of recent periods to use for beta calculation.
                                If None, use all available data.
    
    Returns:
        Tensor of beta coefficients with shape (num_assets,)
    """
    # Default beta is 1.0
    num_assets = asset_returns.shape[1]
    betas = torch.ones(num_assets)
    
    # Use all available data if beta_measurement_period is not specified
    if beta_measurement_period is None or beta_measurement_period > len(asset_returns):
        beta_measurement_period = len(asset_returns)
    
    # Only calculate betas if we have enough data
    if len(asset_returns) > 0:
        # Get the most recent data for beta calculation
        end_idx = len(asset_returns)
        start_idx = end_idx - beta_measurement_period
        
        # Calculate beta for each asset
        for i in range(num_assets):
            asset_i_returns = asset_returns[start_idx:end_idx, i]
            bench_i_returns = benchmark_returns[start_idx:end_idx, 0]
            
            # Calculate covariance and variance for beta
            cov = torch.mean((asset_i_returns - torch.mean(asset_i_returns)) * 
                            (bench_i_returns - torch.mean(bench_i_returns)))
            var = torch.var(bench_i_returns, unbiased=False)
            
            # Avoid division by zero
            if var > 0:
                betas[i] = cov / var

    return betas



"""
Optimize a strategy on historical data
Simulate N prices paths with model fitted to a market
If strategy has a real alpha, it will underperform on the random paths
"""
# TODO: support not only prices as information for allocator

from MLTT.models.price_models import PriceModel
from MLTT.report import Metric
from MLTT.report.metrics import AnnualSharpe
from MLTT import CapitalAllocator
from MLTT.utils import change

from scipy.stats import ttest_1samp

import torch


class RandomPermutationsPriceModel(PriceModel):
    """
    Price model that generates random permutations of the historical data
    """

    def fit(self, historical_prices: torch.Tensor) -> None:
        """
        Just store the historical prices
        
        Args:
            - `historical_prices` (torch.Tensor): Historical price data to be stored
        """
        self.historical_prices = historical_prices

    def _simulate_path(self) -> torch.Tensor:
        """
        Permutate the price changes
        
        Returns:
            torch.Tensor: A simulated price path based on permutation of historical changes
        """
        price_change = change(self.historical_prices)
        # Using torch's random permutation
        idx = torch.randperm(price_change.size(0))
        permuted_price_change = price_change[idx]
        return torch.cumsum(permuted_price_change, dim=0)

    def simulate_paths(self, num_paths: int, *args, **kwargs) -> torch.Tensor:
        """
        Simulate multiple paths of the asset price.

        Args:
            - `num_paths` (int): Number of paths to simulate
            
        Returns:
            torch.Tensor: Tensor of simulated price paths
        """
        paths = [self._simulate_path() for _ in range(num_paths)]
        return torch.stack(paths)
    

def _overfit_pvalue(strategy_metric: float, metric_randoms: list[float]) -> float:
    """
    Test if strategy is likely to be overfited.

    H0: strategy is overfitted
    H1: strategy is not overfitted

    Args:
        - `strategy_metric` (float): The metric value for the strategy
        - `metric_randoms` (list[float]): List of metric values from random simulations
        
    Returns:
        float: p-value of t-test. Lower p-value is better in terms of good fitting.
    """
    # Convert to numpy array for ttest_1samp
    metric_randoms_np = torch.tensor(metric_randoms).cpu().numpy() if isinstance(metric_randoms[0], torch.Tensor) else metric_randoms
    return ttest_1samp(metric_randoms_np, strategy_metric, alternative="greater").pvalue


def _test_n_paths(
    allocator: CapitalAllocator,
    prices: torch.Tensor,
    metric: Metric | None = None,
    price_model: PriceModel | None = None,
    num_paths: int = 100,
) -> float:  # TODO: add asyncronous backtest
    if metric is None:
        metric = AnnualSharpe()
    if price_model is None:
        price_model = RandomPermutationsPriceModel()
    price_model.fit(prices)
    time_steps = prices.shape[0]

    random_prices = price_model.simulate_paths(num_paths=num_paths, time_steps=time_steps)
    # TODO


import torch
import torch.nn.functional as F
from MLTT.utils import ensure_tensor, CACHE_SIZE
from MLTT.cache import conditional_lru_cache


@conditional_lru_cache(maxsize=CACHE_SIZE)
def price_channel(high: torch.Tensor,
                   low: torch.Tensor = None, 
                   lookback: int = 14, 
                   bounds: float = 0.8) -> tuple[torch.Tensor, torch.Tensor]:
    """
    Calculate price channel with upper and lower bounds.
    
    The channel is calculated by finding the highest high and lowest low over the lookback period.
    The width of the channel is adjusted to be bounds% of the midpoint.
    
    Args:
        - `high` (torch.Tensor): High prices tensor of shape (n_samples,) or (n_samples, n_assets)
        - `low` (torch.Tensor): Low prices tensor of shape (n_samples,) or (n_samples, n_assets). If None, uses close prices from high tensor
        - `lookback` (int): Period for calculating channel bounds
        - `bounds` (float): Width of the channel as percentage (0.0-1.0) from midpoint
            e.g. 0.8 means ±40% from midpoint
            
    Returns:
        tuple[torch.Tensor, torch.Tensor]: tuple of (upper_bound, lower_bound) tensors matching input shape
        First lookback-1 values will be NaN
    """
    high, low = ensure_tensor(high, low)
    if low is None:
        low = high
        
    # Handle both single asset and multi-asset cases
    original_shape = high.shape
    if high.dim() == 1:
        high = high.view(-1, 1)
        low = low.view(-1, 1)
    
    n_samples, n_assets = high.shape
    
    # Initialize output tensors with NaN
    resistance = torch.full_like(high, float('nan'))
    support = torch.full_like(low, float('nan'))
    
    # Calculate valid range values using max_pool1d for each asset
    for i in range(n_assets):
        high_pad = high[:, i].view(1, 1, -1)
        low_pad = low[:, i].view(1, 1, -1)
        
        valid_resistance = -F.max_pool1d(-high_pad, kernel_size=lookback, stride=1).view(-1)
        valid_support = F.max_pool1d(low_pad, kernel_size=lookback, stride=1).view(-1)
        
        # Assign valid values starting from lookback-1 index
        resistance[lookback-1:, i] = valid_resistance
        support[lookback-1:, i] = valid_support
    
    # Calculate midpoint and range
    midpoint = (resistance + support) / 2
    channel_range = resistance - support
    
    # Adjust channel width based on bounds parameter
    upper_bound = midpoint + (channel_range * bounds) / 2
    lower_bound = midpoint - (channel_range * bounds) / 2
    
    # Restore original shape if input was 1D
    if len(original_shape) == 1:
        upper_bound = upper_bound.view(-1)
        lower_bound = lower_bound.view(-1)
    
    return upper_bound, lower_bound


from MLTT.validation.walkforward import WalkForward
from MLTT.validation.buckets import BucketWatcher
from MLTT.validation.leak_check import (
    WeightConsistencyChecker, 
    check_random_data_leaks,
    LeakTestResult,
    RandomLeakTester,
)
from MLTT.validation.insample_overfit import RandomPermutationsPriceModel



"""
Module for testing strategies on different buckets by some feature.
For example top 10% volatility, bottom 10% volatility (price, vol of vol etc)
"""
import torch
from typing import Callable
from enum import Enum

from MLTT.allocation.backtesting import backtest, BTResult, backtest_model
from MLTT import CapitalAllocator


class BucketMode(str, Enum):
    """Mode for bucket testing"""
    SUBSET_INPUT = "SUBSET_INPUT"  # Give model only data from bucket subset
    FILTER_OUTPUT = "FILTER_OUTPUT"  # Give model all data but filter weights by bucket


class BucketWatcher:
    def __init__(self, prices: torch.Tensor, prediction_info: torch.Tensor | None = None):
        """
        Args:
            - `prices` (torch.Tensor): 2-dimensional array of log-prices
            - `prediction_info` (torch.Tensor | None): Additional information for model prediction
                Shape: `(batch_size, *n_information)`
        """
        self.prices = prices
        self.prediction_info = prediction_info
        self.feature_indices = None
        self.feature_names = None

    def _validate_feature_values(self, 
                                 feature_values: torch.Tensor | list | float, 
                                 data: torch.Tensor) -> torch.Tensor:
        """Validate and convert feature values to correct tensor format
        
        Args:
            - `feature_values` (torch.Tensor | list | float): Values returned by feature function
            - `data` (torch.Tensor): Original data tensor used for feature calculation
            
        Returns:
            torch.Tensor: Validated 1-dimensional tensor
            
        Raises:
            ValueError: If feature values have incorrect shape or size
        """
        if not isinstance(feature_values, torch.Tensor):
            feature_values = torch.tensor(feature_values)
            
        if feature_values.dim() != 1:
            raise ValueError(
                f"Feature function must return 1-d tensor, got {feature_values.dim()}"
            )
            
        if feature_values.size(0) != data.size(1):
            raise ValueError(
                f"Feature function must return tensor of size {data.size(1)}, got {feature_values.size(0)}"
            )
            
        return feature_values

    def _validate_bucket_state(self) -> None:
        """Validate that buckets were created before backtesting
        
        Raises:
            ValueError: If buckets were not created
        """
        if self.feature_indices is None or self.feature_names is None:
            raise ValueError("call make_buckets before backtest_buckets")

    def _backtest_bucket(
        self,
        indices: torch.Tensor,
        model: CapitalAllocator | None,
        mode: BucketMode,
        precalculated_weights: torch.Tensor | None = None,
        **backtest_kwargs
    ) -> BTResult:
        """Run backtest for a single bucket
        
        Args:
            - `indices` (torch.Tensor): Boolean tensor indicating which assets belong to the bucket
            - `model` (CapitalAllocator | None): Model to use for predictions
            - `mode` (BucketMode): Testing mode
            - `precalculated_weights` (torch.Tensor | None): Pre-calculated weights for FILTER_OUTPUT mode
            - `**backtest_kwargs`: Additional arguments for backtest function
            
        Returns:
            BTResult: Backtest result for the bucket
        """
        bucket_prices = self.prices[:, indices]
        bucket_info = (
            self.prediction_info[:, indices] 
            if self.prediction_info is not None 
            else None
        )

        if mode == BucketMode.SUBSET_INPUT:
            if model is None:
                raise ValueError("Model required for SUBSET_INPUT mode")
            
            return backtest_model(
                model=model,
                prices=bucket_prices,
                prediction_info=bucket_info,
                **backtest_kwargs
            )
        else:  # FILTER_OUTPUT mode
            if precalculated_weights is None:
                raise ValueError("Pre-calculated weights required for FILTER_OUTPUT mode")
                
            # Filter weights by bucket indices
            bucket_weights = precalculated_weights[:, indices]
            
            # Backtest filtered weights
            return backtest(
                weights=bucket_weights,
                prices=bucket_prices,
                **backtest_kwargs
            )

    def make_buckets(
        self,
        feature: Callable | list[Callable],
        quantile: float | list[float] = 0.1
    ) -> tuple[torch.Tensor, list[str]]:
        """
        Separates assets into buckets according to some feature (top 10% and bottom 10%)
        
        Feature function should take a tensor of shape (time_steps, n_assets) and return
        tensor of shape (n_assets,) with feature values for each asset.
        
        saves `feature_names`, `feature_indices` for further use

        Args:
            - `feature` (Callable | list[Callable]): function (or list of functions) that takes
                2-d tensor and returns 1-d tensor with feature values for each asset
            - `quantile` (float | list[float]): quantile for top and bottom buckets.
                (`quantile=0.1` means top 10% and bottom 10% buckets in returned array)

        Returns:
            - torch.Tensor: 2-dimensional array of indices of buckets. Shape: `(n_buckets, n_information)`
            - list[str]: list of feature names with quantile
        """
        features = [feature] if isinstance(feature, Callable) else feature
        quantiles = [quantile] if isinstance(quantile, float) else quantile

        self.feature_indices = []
        self.feature_names = []
        
        data = self.prices if self.prediction_info is None else self.prediction_info
        
        for f in features:
            for q in quantiles:
                # Calculate feature values for all assets at once
                feature_values = self._validate_feature_values(f(data), data)
                
                top_quantile = torch.quantile(feature_values, 1 - q)
                bottom_quantile = torch.quantile(feature_values, q)

                bottom_bucket_idx = feature_values <= bottom_quantile
                top_bucket_idx = feature_values >= top_quantile

                self.feature_names.extend([f'top_{q}_{f.__name__}', f'bottom_{q}_{f.__name__}'])
                self.feature_indices.extend([top_bucket_idx, bottom_bucket_idx])

        self.feature_indices = torch.stack(self.feature_indices, dim=0)
        return self.feature_indices, self.feature_names

    def backtest_buckets(
        self,
        model: CapitalAllocator | None = None,
        mode: BucketMode = BucketMode.SUBSET_INPUT,
        **backtest_kwargs
    ) -> dict[str, BTResult]:
        """Test strategy on different buckets using specified mode.
        
        Args:
            - `model` (CapitalAllocator | None): Model to use for predictions. 
                Required for both modes
            - `mode` (BucketMode): Testing mode
            - `**backtest_kwargs`: Additional arguments for backtest function
                
        Returns:
            dict[str, BTResult]: backtest results for each bucket
            
        Note:
            In FILTER_OUTPUT mode, the function will first run model with full data
            to get weights, then filter those weights for each bucket.
        """
        self._validate_bucket_state()
        
        if model is None:
            raise ValueError("Model is required for both modes")
            
        # Pre-calculate weights for FILTER_OUTPUT mode
        precalculated_weights = None
        if mode == BucketMode.FILTER_OUTPUT:
            input_data = self.prediction_info if self.prediction_info is not None else self.prices
            raw_weights = model(input_data)
            precalculated_weights = torch.roll(raw_weights, 1, dims=0)
            
        return {
            name: self._backtest_bucket(
                indices, 
                model, 
                mode, 
                precalculated_weights=precalculated_weights,
                **backtest_kwargs
            )
            for name, indices in zip(self.feature_names, self.feature_indices)
        }



# TODO: SWITCH TO TIMESCALEDB
from MLTT.data_loading.dumper import dump_data
from MLTT.data_loading.formatter import format_files, sort_dates

import os
import asyncio
import ccxt.async_support as ccxt
import pandas as pd
import numpy as np
from MLTT.utils import EPSILON

def load_and_format(tickers: list[str] | None,
                    interval: str,
                    ask: bool = True,
                    asset_class='spot',
                    path='../../data') -> None:
    """Loads financial data from Binance exchange into local storage and 
        formats it for further analysis.

    Args:
        - `tickers` (list[str]): List of ticker symbols to download data for. 
            Defaults to all available USDT tickers.
        - `interval` (str): Time interval of the data, e.g. '1m', '1h', '1d'.
        - `ask` (bool): Whether to ask for confirmation before downloading 
            new data. Defaults to True.
        - `asset_class` (str): spot, um (for USD-M futures) or 
            cm (for coin-M futures). Defaults to 'spot'.
        - `path` (str): Path to the directory where to dump the data. 
            Defaults to '../../data'.
    """
    if not ask or input("Are you actually want to download new data? (y/n)").lower() == 'y':
        dump_data(tickers,
                  interval,
                  path,
                  asset_class=asset_class,
                  data_type='klines')
        only_format(interval, path)

def only_format(interval, path='../../data', overwrite=False, verbose=False):
    format_files(
        os.path.join(path, 'spot', 'monthly'),
        interval, os.path.join(path, interval), 
        verbose=verbose, overwrite=overwrite
    )
    sort_dates(os.path.join(path, interval))

def save_symbols(symbols, file_path):
    """
    Saves a list of strings (symbols) to a file (file_path).
    Each string is written on a new line in the file.

    Args:
        - `symbols` (list): List of strings to save.
        - `file_path` (str): Path to the file for writing.
    """
    try:
        with open(file_path, 'w') as file:
            for symbol in sorted(symbols):
                file.write(symbol + '\n')
        print(f"Symbols successfully saved to file: {file_path}")
    except Exception as e:
        print(f"Error saving symbols to file: {e}")

def load_symbols(file_path):
    """
    Loads a list of strings from a file (file_path).
    Each line in the file is treated as a separate element in the list.

    Args:
        - `file_path` (str): Path to the file for reading.

    Returns:
        list: List of strings loaded from the file.
    """
    try:
        with open(file_path, 'r') as file:
            symbols = [line.strip() for line in file.readlines()]
        print(f"Successfully loaded {len(symbols)} symbols from file: {file_path}")
        return sorted(symbols)
    except Exception as e:
        print(f"Error loading symbols from file: {e}")
        return []

async def fetch_ohlcv(exchange, sym, timeframe):
    try:
        ohlcv = await exchange.fetch_ohlcv(sym, timeframe, limit=1000)
        df = pd.DataFrame(ohlcv, columns=['timestamp', 'open', 'high', 'low', 'close', 'volume'])
        df['timestamp'] = pd.to_datetime(df['timestamp'], unit='ms')
        df.set_index('timestamp', inplace=True)
        return sym, df
    except Exception as e:
        print(f"Error fetching data for {sym}: {e}")
        return sym, None

async def download_ohlcv(exchange_name: str, symbol: str | list[str], timeframe: str):
    """
    Asynchronously downloads OHLCV data from the specified exchange for one or multiple pairs.

    Args:
        - `exchange_name` (str): Name of the exchange (e.g. 'binance', 'kraken').
        - `symbol` (str | list[str]): Trading pair or list of pairs (e.g. 'BTC/USDT' or ['BTC/USDT', 'ETH/USDT']).
        - `timeframe` (str): Time interval (e.g. '1m', '1h', '1d').

    Returns:
        dict: Dictionary containing OHLCV data for each pair.
    """
    exchange_class = getattr(ccxt, exchange_name)
    exchange = exchange_class({'enableRateLimit': True})

    try:
        if isinstance(symbol, str):
            symbol = [symbol]

        tasks = [fetch_ohlcv(exchange, sym, timeframe) for sym in symbol]
        results = await asyncio.gather(*tasks)

        data = {sym: df for sym, df in results if df is not None}

    finally:
        await exchange.close()

    return data

def ccxt_ohlcv(exchange_name: str, symbol: str | list[str], timeframe: str):
    """
    Downloads OHLCV data from the specified exchange using CCXT.

    Args:
        - `exchange_name` (str): Name of the exchange (e.g. 'binance', 'kraken').
        - `symbol` (str | list[str]): Trading pair or list of pairs.
        - `timeframe` (str): Time interval.

    Returns:
        dict: Dictionary containing OHLCV data for each pair.
    """
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(download_ohlcv(exchange_name, symbol, timeframe))


def _convert_timestamp(ts: int | float) -> pd.Timestamp:
    """Convert timestamp with automatic unit detection"""
    try:
        return pd.to_datetime(ts, unit='ms')
    except pd.errors.OutOfBoundsDatetime:
        return pd.to_datetime(ts, unit='ns')

def check_linearity(timestamps: np.ndarray, tolerance: float = 0.05) -> bool:
    """Check if timestamps have constant intervals within given tolerance."""
    if len(timestamps) < 2:
        return True

    deltas = np.diff(timestamps)
    
    max_step = np.max(deltas)
    min_step = np.min(deltas)
    is_linear = max_step - min_step <= 2*tolerance*np.median(deltas)
    return bool(is_linear)

def check_klines_timeline(
        klines: np.ndarray, 
        deep: bool = True, 
        verbose: bool = True,
        linear: bool = True,
        tolerance: float = 0.05
    ) -> bool:
    if deep:
        timestamps = klines[:, :, 0]
        reference = timestamps[0]
        timeline_passed = all(np.array_equal(ts, reference) for ts in timestamps)
        passed = timeline_passed

        linear_passed = True
        if timeline_passed and linear:
            linear_passed = check_linearity(reference, tolerance)
            passed = passed and linear_passed
    else:
        corrected_starts = klines[:, 0, 0]
        corrected_ends = klines[:, -1, 0]
        passed = len(np.unique(corrected_starts)) == len(np.unique(corrected_ends)) == 1

    if verbose:
        if deep:
            timeline_status = "passed ✅" if timeline_passed else "failed ❌"
            print(f"Same timeline check {timeline_status}")

            if linear and timeline_passed:
                lin_status = "passed ✅" if linear_passed else "failed ❌"
                print(f"Linear timeline check {lin_status}")
        else:
            passed_str = "passed ✅" if passed else "failed ❌"
            print(f"Same timeline check {passed_str}")

        if passed:
            try:
                start_ts = klines[0, 0, 0]
                end_ts = klines[0, -1, 0]
                start_date = _convert_timestamp(start_ts)
                end_date = _convert_timestamp(end_ts)

                print(f"\nStart date: {start_date.strftime('%d %b %Y %H:%M:%S.%f')[:-3]}")
                print(f"End date: {end_date.strftime('%d %b %Y %H:%M:%S.%f')[:-3]}")
            except Exception as e:
                print(f"\nError converting timestamps: {e}")
                print(f"Raw start timestamp: {start_ts}")
                print(f"Raw end timestamp: {end_ts}")

    return passed



from MLTT.allocation import BaseAllocator
from MLTT.utils import to_weights_matrix

from torch import Tensor, sign, roll


def sign_model(cls):
    """
    A decorator that converts model's signals to -1, 0, or 1.
    Modifies the predict method of any allocator model to output sign values.
    
    Usage:
        @sign_model
        class MyModel(BaseAllocator):
            ...
    """
    original_predict = cls.predict
    
    def new_predict(*args, **kwargs) -> Tensor:
        # Get predictions from original method
        predictions = original_predict(*args, **kwargs)
        # Convert to -1, 0, 1 signals
        signals = sign(predictions)
        return to_weights_matrix(signals)
        
    cls.predict = new_predict
    return cls

@sign_model
class LeakModel(BaseAllocator):
    """
    A model that "leaks" future information by looking one step ahead.
    For each asset:
    - Takes long position if price will increase next day
    - Takes short position if price will decrease next day
    
    Warning: This model uses future information and should only be used for research/testing purposes.
    """

    def __init__(self) -> None:
        super().__init__(num_observations=1)

    def predict(self, x: Tensor) -> Tensor:
        # Calculate next day's returns by shifting prices back
        next_prices = roll(x, shifts=-1, dims=0)
        future_returns = (next_prices - x)
        return future_returns



from abc import ABC
from torch import Tensor, mean

from allocation_o2 import CapitalAllocator # type: ignore


class BaseAllocator(CapitalAllocator, ABC):
    def __init__(self,
                 num_observations: int = 0) -> None:
        self.num_observations = num_observations

    @property
    def min_observations(self) -> int:
        return self.num_observations


class PriceAwareAllocator(CapitalAllocator, ABC):
    def get_prices_matrix(self) -> Tensor:
        ...


class BlendingModel(CapitalAllocator):
    def __init__(self, models: list[CapitalAllocator], weights: Tensor) -> None:
        self.models = models
        self.weights = weights

    @property
    def min_observations(self) -> int:
        return max([model.min_observations for model in self.models])

    def predict(self, x: Tensor) -> Tensor:
        predictions = [model.predict(x) for model in self.models]
        return mean(predictions, dim=0, weights=self.weights)  # TODO: check axis



#TODO: make geometric brownian motion usual brownian motion (log-scale)
"""
Interface and implementations for price models that can be calibrated to market data
and used for price simulation.
"""
from abc import ABC, abstractmethod
from scipy.optimize import minimize

import torch
from MLTT.utils import change

class PriceModel(ABC):
    """
    Base class for all price models.
    """

    def __init__(self):
        self.generated_paths = None
        self.time_steps_generated = None

    @abstractmethod
    def simulate_paths(self, num_paths: int, time_steps: int) -> torch.Tensor:
        """
        Simulate multiple paths of the asset price.
        
        Args:
            - `num_paths` (int): Number of paths to simulate
            - `time_steps` (int): Number of time steps in simulation
            
        Returns:
            - torch.Tensor: Simulated paths as a tensor of shape (time_steps + 1, num_paths)
        """
        pass

    @abstractmethod
    def fit(self, historical_prices: torch.Tensor) -> None:
        """
        Fit the model parameters based on historical price data.
        
        Args:
            - `historical_prices` (torch.Tensor): Historical price data as a tensor
        """
        pass


class GBMModel(PriceModel):
    """
    Classical Geometric Brownian Motion model with constant drift and volatility.
    """
    
    def __init__(self, initial_price, drift=None, volatility=None):
        """
        Initialize Geometric Brownian Motion model.
        
        Args:
            - `initial_price` (float): Initial price of the asset
            - `drift` (float | None): Expected return (drift) of the asset per time step
            - `volatility` (float | None): Volatility of the asset per time step
        """
        super().__init__()
        self.initial_price = initial_price
        self.drift = drift
        self.volatility = volatility

    def simulate_paths(self, num_paths: int, time_steps: int) -> torch.Tensor:
        """
        Simulate multiple price paths using Geometric Brownian Motion.
        
        Args:
            - `num_paths` (int): Number of paths to simulate
            - `time_steps` (int): Number of time steps in simulation
            
        Returns:
            - torch.Tensor: Simulated paths as a tensor of shape (time_steps + 1, num_paths)
        """
        paths = torch.zeros((time_steps + 1, num_paths))
        paths[0] = self.initial_price

        for t in range(1, time_steps + 1):
            random_shocks = torch.randn(num_paths)
            paths[t] = paths[t - 1] * torch.exp(
                (self.drift - 0.5 * self.volatility ** 2) +
                self.volatility * random_shocks
            )
            
        self.generated_paths = paths
        self.time_steps_generated = torch.arange(time_steps + 1)
        return paths

    def fit(self, historical_prices: torch.Tensor) -> None:
        """
        Fit drift and volatility parameters using historical prices.
        
        Args:
            - `historical_prices` (torch.Tensor): Historical price data
        """
        log_returns = torch.diff(torch.log(historical_prices))
        self.drift = torch.mean(log_returns)
        self.volatility = torch.std(log_returns)


class HestonModel(PriceModel):  # TODO: Check if Heston model is correct and fits to the market data
    """
    Heston stochastic volatility model.
    """
    
    def __init__(self, initial_price, v0, kappa, theta, sigma, rho):
        """
        Initialize Heston stochastic volatility model.
        
        Args:
            - `initial_price` (float): Initial price of the asset
            - `v0` (float): Initial variance
            - `kappa` (float): Rate of mean reversion
            - `theta` (float): Long-term variance
            - `sigma` (float): Volatility of volatility
            - `rho` (float): Correlation between asset returns and variance
        """
        super().__init__()
        self.initial_price = initial_price
        self.v0 = v0
        self.kappa = kappa
        self.theta = theta
        self.sigma = sigma
        self.rho = rho

    def simulate_paths(self, num_paths: int, time_steps: int) -> torch.Tensor:
        """
        Simulate multiple price paths using Heston stochastic volatility model.
        
        Args:
            - `num_paths` (int): Number of paths to simulate
            - `time_steps` (int): Number of time steps in simulation
            
        Returns:
            - torch.Tensor: Simulated paths as a tensor of shape (time_steps + 1, num_paths)
        """
        dt = 1 / time_steps
        S = torch.full((time_steps + 1, num_paths), self.initial_price)
        v = torch.full((time_steps + 1, num_paths), self.v0)

        for t in range(1, time_steps + 1):
            Z = torch.randn(num_paths, 2)
            Z = torch.cat([Z, -Z], dim=1)
            
            v[t] = torch.maximum(
                v[t - 1] + 
                self.kappa * (self.theta - v[t - 1]) * dt + 
                self.sigma * torch.sqrt(v[t - 1] * dt) * Z[:, 1], 
                0
            )
            
            exponent = (self.kappa - 0.5 * v[t - 1]) * dt + torch.sqrt(v[t - 1] * dt) * Z[:, 0]
            exponent = torch.clip(exponent, -700, 700)
            S[t] = S[t - 1] * torch.exp(exponent)
        
        self.generated_paths = S
        self.time_steps_generated = torch.arange(time_steps + 1)
        return S

    def _log_likelihood(self, params, historical_prices):
        """
        Helper method for MLE fitting.
        
        Args:
            - `params` (list): Model parameters [v0, kappa, theta, sigma, rho]
            - `historical_prices` (torch.Tensor): Historical price data
            
        Returns:
            - float: Negative log-likelihood (for minimization)
        """
        v0, kappa, theta, sigma, rho = params
        self.v0, self.kappa, self.theta, self.sigma, self.rho = params
        
        simulated_path = self._simulate_calibration_path(historical_prices)
        
        log_returns_hist = torch.diff(torch.log(historical_prices))
        log_returns_sim = torch.diff(torch.log(simulated_path))
        
        residuals = log_returns_hist - log_returns_sim
        log_likelihood = -0.5 * torch.sum(residuals ** 2)
        
        return -log_likelihood
    
    def _simulate_calibration_path(self, historical_prices):
        """
        Simulate a single path for calibration purposes.
        
        Args:
            - `historical_prices` (torch.Tensor): Historical price data to match
            
        Returns:
            - torch.Tensor: Simulated price path for calibration
        """
        time_steps = len(historical_prices)
        dt = 1 / time_steps
        path = torch.zeros(time_steps)
        path[0] = historical_prices[0]
        v = torch.zeros(time_steps)
        v[0] = self.v0

        Z = torch.randn(time_steps-1, 2)
        Z = torch.cat([Z, -Z], dim=1)

        for t in range(1, time_steps):
            v[t] = torch.maximum(
                v[t-1] + 
                self.kappa * (self.theta - v[t-1]) * dt + 
                self.sigma * torch.sqrt(v[t-1] * dt) * Z[t-1, 1], 
                0
            )
            
            exponent = (self.kappa - 0.5 * v[t-1]) * dt + torch.sqrt(v[t-1] * dt) * Z[t-1, 0]
            exponent = torch.clip(exponent, -700, 700)
            path[t] = historical_prices[t-1] * torch.exp(exponent)

        return path

    def fit(self, historical_prices: torch.Tensor) -> None:
        """
        Fit Heston model parameters using maximum likelihood estimation.
        
        Args:
            - `historical_prices` (torch.Tensor): Historical price data
        """
        initial_params = [self.v0, self.kappa, self.theta, self.sigma, self.rho]
        
        bounds = [
            (1e-6, None),  # v0: variance > 0
            (1e-6, None),  # kappa: mean-reversion speed > 0
            (1e-6, None),  # theta: long-term variance > 0
            (1e-6, None),  # sigma: vol of vol > 0
            (-1, 1)        # rho: correlation in [-1, 1]
        ]

        result = minimize(
            self._log_likelihood, 
            initial_params,
            args=(historical_prices,),
            bounds=bounds, 
            method='L-BFGS-B'
        )
        
        if result.success:
            self.v0, self.kappa, self.theta, self.sigma, self.rho = result.x
        else:
            raise RuntimeError("Failed to fit Heston model parameters!")

            
class HistoricalPriceMotion(PriceModel):
    """
    Model that simulates price paths by sampling from historical price changes.
    """
    
    def __init__(self, initial_price, historical_prices=None):
        """
        Initialize historical price motion model.
        
        Args:
            - `initial_price` (float): Initial price of the asset
            - `historical_prices` (torch.Tensor | None): Historical price data to fit the model
        """
        super().__init__()
        self.initial_price = initial_price
        self.changes = None
        if historical_prices is not None:
            self.fit(historical_prices)
    
    def fit(self, historical_prices):
        """
        Calculate historical price changes for sampling.
        
        Args:
            - `historical_prices` (torch.Tensor): Historical price data
        """
        self.changes = change(historical_prices)
    
    def simulate_paths(self, num_paths, time_steps):
        """
        Simulate multiple paths by sampling historical changes.
        
        Args:
            - `num_paths` (int): Number of paths to simulate
            - `time_steps` (int): Number of time steps to simulate
            
        Returns:
            - torch.Tensor: Simulated paths tensor of shape (time_steps + 1, num_paths)
        """
        if self.changes is None:
            raise ValueError("Model must be fit before simulating paths")
            
        # Sample changes with replacement
        changes = torch.randint(
            low=0,
            high=len(self.changes),
            size=(time_steps, num_paths),
            replace=True
        )
        
        # Calculate cumulative product of (1 + changes)
        paths = self.initial_price * torch.cumprod(1 + changes, axis=0)
        
        # Add initial price as first row
        paths = torch.vstack([
            torch.full((1, num_paths), self.initial_price),
            paths
        ])
        
        self.generated_paths = paths
        self.time_steps_generated = torch.arange(time_steps + 1)
        
        return paths



"""
Functions for generating and optimizing alpha factors.
"""
import os
from dataclasses import dataclass

import pandas as pd
from torch import Tensor

from MLTT.alphas.base import AlphaFactor
from MLTT.alphas.data import AlphaData
from MLTT.alphas.generation import generate_random_alpha
from MLTT.alphas.performance import compute_ic
from MLTT.alphas.visualization import visualize_alpha, print_alpha_tree


@dataclass
class AlphaResult:
    """Container for alpha generation result."""
    
    alpha: AlphaFactor
    ic: float
    stats: dict[str, float]
    values: Tensor

def generate_and_evaluate_alphas(
    data: AlphaData,
    returns: Tensor,
    n_alphas: int = 50,
    max_depth: int = 5,
    window_range: tuple[int, int] = (12, 48),
    save_dir: str | None = None
) -> list[AlphaResult]:
    """Generate and evaluate multiple alpha factors.
    
    Args:
        - `data` (AlphaData): Input data
        - `returns` (Tensor): Forward returns for IC calculation
        - `n_alphas` (int): Number of alphas to generate
        - `max_depth` (int): Maximum depth of alpha trees
        - `window_range` (tuple[int, int]): Range for window parameters
        - `save_dir` (str | None): Directory to save visualizations
        
    Returns:
        list[AlphaResult]: List of alpha results, sorted by absolute IC value
    """
    results = []
    
    print(f"\nGenerating {n_alphas} alpha factors...")
    for i in range(n_alphas):
        print(f"\nAlpha {i+1}/{n_alphas}")
        
        # Generate alpha
        alpha = generate_random_alpha(
            data=data,
            max_depth=max_depth,
            window_range=window_range
        )
        
        # Print formula
        print("Formula:", alpha.to_string())
        
        # Compute values
        values = alpha.compute(data.values)
        
        # Compute IC
        ic = compute_ic(values, returns)
        print(f"IC: {ic:.4f}")
        
        # Save result
        results.append(AlphaResult(alpha, ic, {}, values))
        
        # Save visualization if requested
        if save_dir is not None:
            os.makedirs(save_dir, exist_ok=True)
            visualize_alpha(
                alpha,
                filename=os.path.join(save_dir, f"alpha_{i+1}"),
                format="png",
                view=False
            )
    
    # Sort by absolute IC value
    results.sort(key=lambda x: abs(x.ic), reverse=True)
    
    return results


def print_alpha_results(
    results: list[AlphaResult],
    n_best: int = 10,
    save_path: str | None = None
) -> None:
    """Print results of alpha generation.
    
    Args:
        - `results` (list[AlphaResult]): List of alpha results
        - `n_best` (int): Number of best alphas to show
        - `save_path` (str | None): Path to save results CSV
    """
    print(f"\nTop {n_best} alphas by IC:")
    for i, result in enumerate(results[:n_best], 1):
        print(f"\n{i}. IC: {result.ic:.4f}")
        print("Formula:", result.alpha.to_string())
        print("Statistics:")
        for name, value in result.stats.items():
            print(f"  {name}: {value:.4f}")
        print("\nTree structure:")
        print_alpha_tree(result.alpha)
    
    if save_path:
        # Create DataFrame with results
        df = pd.DataFrame([
            {
                'rank': i + 1,
                'ic': r.ic,
                'formula': r.alpha.to_string(),
                **{f'stat_{k}': v for k, v in r.stats.items()}
            }
            for i, r in enumerate(results)
        ])
        
        # Save to CSV
        df.to_csv(save_path, index=False)
        print(f"\nResults saved to {save_path}")


def find_best_alpha(
    data: AlphaData,
    returns: Tensor,
    n_alphas: int = 50,
    max_depth: int = 5,
    window_range: tuple[int, int] = (12, 48),
    save_dir: str | None = "alpha_results",
    n_best: int = 10
) -> AlphaResult:
    """Generate multiple alphas and find the best one.
    
    Args:
        - `data` (AlphaData): Input data
        - `returns` (Tensor): Forward returns for IC calculation
        - `n_alphas` (int): Number of alphas to generate
        - `max_depth` (int): Maximum depth of alpha trees
        - `window_range` (tuple[int, int]): Range for window parameters
        - `save_dir` (str | None): Directory to save results
        - `n_best` (int): Number of best alphas to show
        
    Returns:
        AlphaResult: Best alpha result
    """
    # Create save directory
    if save_dir:
        os.makedirs(save_dir, exist_ok=True)
        vis_dir = os.path.join(save_dir, "visualizations")
        os.makedirs(vis_dir, exist_ok=True)
    else:
        vis_dir = None
    
    # Generate and evaluate alphas
    results = generate_and_evaluate_alphas(
        data=data,
        returns=returns,
        n_alphas=n_alphas,
        max_depth=max_depth,
        window_range=window_range,
        save_dir=vis_dir
    )
    
    # Print and save results
    if save_dir:
        csv_path = os.path.join(save_dir, "results.csv")
    else:
        csv_path = None
    
    print_alpha_results(results, n_best=n_best, save_path=csv_path)
    
    return results[0]  # Return best alpha


# TODO: SWITCH TO TIMESCALEDB
from datetime import datetime
import os
import json
import csv
from warnings import warn

import numpy as np
from binance_historical_data import BinanceDataDumper
from tqdm.auto import tqdm


BEGIN_DATE = datetime(2016, 1, 1).date()
END_DATE   = datetime.today().date()
CACHE_FILE = 'line_counts_cache.json'


def dump_data(tickers: list[str],
              interval: str,
              path: str,
              asset_class="spot",
              data_type="klines") -> None:
    """
    Dumps historical market data from Binance exchange and saves it locally.

    Args:
        - `tickers` (list[str]): List of trading pairs to download 
            (e.g. ['BTCUSDT', 'ETHUSDT'])
        - `interval` (str): Timeframe of the data ('1s', '1m', '1h', '1d', etc.)
        - `path` (str): Directory path where to save the downloaded data
        - `asset_class` (str, optional): Type of market data:
            - 'spot' for spot markets
            - 'um' for USD-M futures
            - 'cm' for COIN-M futures
            Defaults to "spot".
        - `data_type` (str, optional): Type of data to download:
            - 'klines' for candlestick data
            - 'aggTrades' for aggregated trades
            - 'trades' for raw trades
            Defaults to "klines".
    """
    data_dumper = BinanceDataDumper(
        path_dir_where_to_dump=path,
        asset_class=asset_class,
        data_type=data_type,
        data_frequency=interval,
    )

    data_dumper.dump_data(
        tickers=tickers,
        date_start=BEGIN_DATE,
        date_end=END_DATE,
        is_to_update_existing=False,
    )

def count_lines(file_path: str) -> int:
    """
    Counts number of lines in a file.

    Args:
        - `file_path` (str): Path to the file

    Returns:
        - `int`: Number of lines in the file
    """
    counter = 0

    with open(file_path, 'r') as f:
        for _ in f.readlines():
            counter += 1
    return counter

def min_lines(files: list[str], update_cache=True) -> tuple[int, dict[str, int]]:
    """
    Finds minimum number of lines across multiple files.
    Uses caching to avoid recounting files that haven't changed.

    Args:
        - `files` (list[str]): List of file paths to check
        - `update_cache` (bool, optional): Whether to update the cache. 
            Defaults to True.

    Returns:
        - `tuple[int, dict[str, int]]`: Tuple containing:
            - Minimum number of lines across all files
            - Dictionary mapping file paths to their line counts
    """
    if update_cache:
        cache = {}
    else:
        cache = _load_cache(CACHE_FILE)
    counts = _get_counts(files, cache)
    _save_cache(CACHE_FILE, counts)

    min_num = _get_min_count(counts)
    return min_num, counts


def _load_cache(cache_file: str) -> dict:
    try:
        with open(cache_file, 'r') as f:
            return json.load(f)
    except FileNotFoundError:
        return {}


def _get_counts(files: list[str], cache: dict) -> dict:
    print("Counting lines in files...")
    counts = {file: cache.get(file, 0) for file in files}
    for file in tqdm(files):
        if file not in cache:
            counts[file] = count_lines(file)
            cache[file] = counts[file]
    return counts


def _save_cache(cache_file: str, cache: dict) -> None:
    with open(cache_file, 'w') as f:
        json.dump(cache, f)


def _get_min_count(counts: dict) -> int:
    return min(counts.values())


def load_klines_from_file(file_path: str, load_cols: list[int], skiprows: int) -> np.ndarray:
    """
    Loads specific columns from a CSV file containing market data.

    Args:
        - `file_path` (str): Path to the CSV file
        - `load_cols` (list[int]): List of column indices to load
        - `skiprows` (int): Number of rows to skip from the beginning

    Returns:
        - `np.ndarray`: Array containing the requested columns
    """
    data = np.loadtxt(
        file_path,
        delimiter=',',
        dtype=np.float64,
        usecols=load_cols,
        skiprows=skiprows,
    )
    return data


def _load_min_csv_files(files: list[str], 
                        load_cols: list[int],
                        update_cache=True,
                        equals=True) -> np.ndarray | list[np.ndarray]:
    """
    Loads multiple CSV files and ensures they have consistent dimensions.

    Args:
        - `files` (list[str]): List of CSV file paths to load
        - `load_cols` (list[int]): Column indices to load from each file
        - `update_cache` (bool, optional): Whether to update line count cache. 
            Defaults to True.
        - `equals` (bool, optional): If True, ensures all loaded arrays have same length. 
            Defaults to True.

    Returns:
        - `np.ndarray | list[np.ndarray]`: Either:
            - Single numpy array with shape (n_files, min_length, n_columns) if equals=True
            - List of arrays with potentially different lengths if equals=False
    """
    min_len, lines = min_lines(files, update_cache=update_cache)

    n_columns = None
    for file_path in files:
        with open(file_path, 'r') as f:
            header = next(csv.reader(f))
            if n_columns is None:
                n_columns = len(header)
            elif len(header) != n_columns:
                raise ValueError(f"Different number of columns in file {file_path}")

    print("Loading files...")

    shape = (len(files), min_len-1, len(load_cols))

    if equals:
        values = np.empty(shape, dtype=np.float64)
    else:
        values = [[] for _ in range(len(files))]

    for i, file_path in enumerate(tqdm(files)):
        skiprows = lines[file_path] - min_len + 1 if equals else 1
        data = load_klines_from_file(file_path, load_cols, skiprows)
        values[i] = data

    return values

def load_csv_files_from_directory(directory: str,
                                  pairs_to_load: list[str] | None = None,
                                  load_cols: list[int] | None = None,
                                  update_cache: bool = True,
                                  equals: bool = True,
                                  close_only: bool | None = None) -> tuple[np.ndarray | list[np.ndarray], list[str]]:
    """
    Loads market data from CSV files in a directory.

    Args:
        - `directory` (str): Directory containing the CSV files
        - `pairs_to_load` (list[str] | None, optional): Specific trading pairs to load. 
            If None, loads all. 
        - `load_cols` (list[int] | None, optional): Specific columns to load from each file.
            Common columns are:
            - 0: OpenTime
            - 1: Open price
            - 2: High price
            - 3: Low price
            - 4: Close price
            - 5: Volume
            - 6: Close time
            - 7: Quote volume
            - 8: Number of trades
            - 9: Taker buy base volume
            - 10: Taker buy quote volume
            If None, loads all columns.
        - `update_cache` (bool, optional): Whether to update line count cache. 
            Defaults to True.
        - `equals` (bool, optional): If True, ensures all loaded data has same length. 
            Defaults to True.

    Returns:
        - `tuple[np.ndarray | list[np.ndarray], list[str]]`: Tuple containing:
            - Array or list of arrays with loaded data
            - List of successfully loaded trading pairs
    """
    if close_only:
        load_cols = [4]
        warn("close_only is deprecated. Use load_cols instead.")

    pairs_to_load = sorted(pairs_to_load) if pairs_to_load else None
    files = sorted(os.listdir(directory))

    file_names = []
    loaded = []
    try:
        for filename in files:
            if not filename.endswith(".csv"):
                continue
            pair = filename.split('_')[0]
            if pairs_to_load and pair not in pairs_to_load:
                continue
            file_names.append(filename)
            loaded.append(pair)

        paths = [os.path.join(directory, filename) for filename in file_names]
        return _load_min_csv_files(paths, load_cols, update_cache, equals), loaded
    except KeyboardInterrupt:
        # Return partial results if keyboard interrupt occurs
        if file_names:  # If we have any loaded files
            paths = [os.path.join(directory, filename) for filename in file_names]
            return _load_min_csv_files(paths, load_cols, update_cache, equals), loaded
        raise  # Re-raise if no files were loaded



from torch import Tensor
from MLTT.allocation import CapitalAllocator
from MLTT.cache import conditional_lru_cache
from MLTT.utils import CACHE_SIZE
import copy


@conditional_lru_cache(maxsize=CACHE_SIZE)
def treshold_change(allocator: CapitalAllocator, threshold: float = 0.2, deepcopy: bool = True) -> CapitalAllocator:
    """
    Decorator for CapitalAllocator that prevents allocation changes unless they exceed a threshold.
    
    Args:
        - `allocator` (CapitalAllocator): The capital allocator to decorate. Stays unchanged. Operates with tensors.
        - `threshold` (float): The minimum sum of absolute differences required to change the allocation
        - `deepcopy` (bool): Whether to deepcopy the allocator. If False, the allocator will be modified in place.
        
    Returns:
        - CapitalAllocator: A decorated CapitalAllocator with threshold-based change prevention
    """
    # Create a copy if requested to avoid modifying original
    decorated_alloc = copy.deepcopy(allocator) if deepcopy else allocator
    
    # Preserve original predict method before any modifications
    original_predict = decorated_alloc.predict
    
    def predict_with_threshold(x: Tensor) -> Tensor:
        # Use preserved original implementation, not current method
        base_allocation = original_predict(x)
        last_allocation = None

        for i in range(base_allocation.shape[0]):
            if last_allocation is None:
                last_allocation = base_allocation[i]
                continue
            
            if (base_allocation[i] - last_allocation).abs().sum() < threshold:
                base_allocation[i] = last_allocation
            else:
                last_allocation = base_allocation[i]

        return base_allocation
    
    decorated_alloc.predict = predict_with_threshold
    return decorated_alloc



import torch
import torch.nn.functional as F
from MLTT.allocation import BaseAllocator
from MLTT.allocation.allocators import PriceAwareAllocator
from MLTT.technical.non_trivial import price_channel
from MLTT.utils import multivariate_barrier, to_weights_matrix, change, EPSILON


def get_price_data(model: BaseAllocator, x: torch.Tensor) -> torch.Tensor:
    """
    Get price data from a model or input tensor.
    Uses PriceAwareAllocator interface if model supports it.
    
    Args:
        - `model` (BaseAllocator): Model that may implement PriceAwareAllocator
        - `x` (torch.Tensor): Input tensor to use as fallback price data
        
    Returns:
        - torch.Tensor: Price tensor
    """
    # Check if model provides price data through PriceAwareAllocator interface
    if isinstance(model, PriceAwareAllocator):
        # Get price matrix from model
        price_matrix = model.get_prices_matrix()
        if price_matrix is not None:
            # Convert to tensor if it's numpy array
            if not isinstance(price_matrix, torch.Tensor):
                price_matrix = torch.tensor(price_matrix, dtype=torch.float32)
            return price_matrix
            
    # Fallback to using input data as prices
    return x


def calculate_asset_volatility(prices: torch.Tensor, lookback_period: int) -> torch.Tensor:
    """
    Calculate historical volatility for each asset
    
    Args:
        - `prices` (torch.Tensor): Price tensor with shape (time_steps, num_assets)
        - `lookback_period` (int): Number of periods to use for volatility calculation
        
    Returns:
        - torch.Tensor: Tensor of volatilities with shape (num_assets,)
    """
    # Need at least 2 observations to calculate returns
    if prices.shape[0] < 2:
        return torch.ones(prices.shape[1])
        
    # Calculate returns - prices are already log prices, so use change
    returns = change(prices, lag=1)
    
    # Use only the most recent lookback_period returns
    if returns.shape[0] > lookback_period:
        recent_returns = returns[-lookback_period:]
    else:
        recent_returns = returns
    
    # Calculate standard deviation of returns
    vol = torch.std(recent_returns, dim=0, unbiased=False)
    
    # Replace zeros with a small value to avoid division by zero
    vol = torch.where(vol > 0, vol, torch.tensor(EPSILON))
    
    return vol


def calculate_portfolio_volatility(weights: torch.Tensor, returns: torch.Tensor) -> float:
    """
    Estimate portfolio volatility based on historical returns and weights
    
    Args:
        - `weights` (torch.Tensor): Weight tensor with shape (num_assets,) or (time_steps, num_assets)
        - `returns` (torch.Tensor): Returns tensor with shape (time_steps, num_assets)
        
    Returns:
        - float: Portfolio volatility as a scalar
    """
    # Ensure weights is a vector (last weight vector if it's a matrix)
    if weights.dim() > 1:
        weights = weights[-1]
        
    # Calculate weighted returns
    weighted_returns = returns @ weights
    
    # Calculate volatility
    portfolio_vol = torch.std(weighted_returns)
    
    return portfolio_vol


class TimeSeriesMomentumModel(BaseAllocator):
    """
    Time Series Momentum strategy implementation.
    Based on the research from "Time Series Momentum" by Moskowitz, Ooi and Pedersen.
    
    The strategy:
    1. Uses past returns to predict future returns
    2. Can incorporate roll returns for futures
    3. Supports position scaling over the holding period
    4. Optional mean reversion filter
    """
    
    def __init__(
        self,
        lookback: int,
        holding_period: int,
        threshold: float = 0.0,
        use_roll_returns: bool = False,
        scale_positions: bool = True,
        use_mean_reversion: bool = False,
        mean_reversion_lookback: int | None = None,
    ):
        """
        Args:
            - `lookback` (int): Number of days to look back for momentum calculation
            - `holding_period` (int): Number of days to hold positions
            - `threshold` (float): Minimum return threshold for taking positions
            - `use_roll_returns` (bool): Whether to use roll returns instead of total returns
            - `scale_positions` (bool): Whether to scale positions over the holding period
            - `use_mean_reversion` (bool): Whether to add mean reversion filter
            - `mean_reversion_lookback` (int | None): Lookback period for mean reversion filter
        """
        self.lookback = lookback
        self.holding_period = holding_period
        self.threshold = threshold
        self.use_roll_returns = use_roll_returns
        self.scale_positions = scale_positions
        self.use_mean_reversion = use_mean_reversion
        self.mean_reversion_lookback = mean_reversion_lookback or (lookback + 10)
        
        # Need enough data for calculations
        super().__init__(num_observations=max(lookback, mean_reversion_lookback or 0))

    def _calculate_returns(self, prices: torch.Tensor) -> torch.Tensor:
        """
        Calculate returns based on configuration
        
        Args:
            - `prices` (torch.Tensor): Price tensor to calculate returns from
            
        Returns:
            - torch.Tensor: Calculated returns
        """
        # Ensure prices is 2D
        if prices.dim() == 1:
            prices = prices.unsqueeze(-1)
            
        if self.use_roll_returns:
            # For futures: calculate roll returns
            # This is a simplified version - in practice you'd need actual futures data
            returns = change(prices, lag=1)
            roll_returns = F.avg_pool1d(
                returns.unsqueeze(0),  # Add batch dimension
                kernel_size=252,
                stride=1,
                padding=251
            ).squeeze(0)  # Remove batch dimension
            return F.pad(roll_returns, (0, 0, 0, 1))  # Pad time dimension
        else:
            # Regular returns over lookback period
            return change(prices, lag=self.lookback)

    def _generate_signals(self, returns: torch.Tensor) -> torch.Tensor:
        """
        Generate trading signals from returns
        
        Args:
            - `returns` (torch.Tensor): Returns tensor to generate signals from
            
        Returns:
            - torch.Tensor: Generated trading signals (-1, 0, 1)
        """
        signals = torch.zeros_like(returns)
        
        # Long signal when return > threshold
        signals[returns > self.threshold] = 1
        
        # Short signal when return < -threshold
        signals[returns < -self.threshold] = -1
        
        return signals

    def _apply_mean_reversion_filter(
        self, 
        signals: torch.Tensor, 
        prices: torch.Tensor
    ) -> torch.Tensor:
        """
        Apply mean reversion filter to signals
        
        Args:
            - `signals` (torch.Tensor): Trading signals to filter
            - `prices` (torch.Tensor): Price tensor for mean reversion calculation
            
        Returns:
            - torch.Tensor: Filtered trading signals
        """
        if not self.use_mean_reversion:
            return signals
            
        # Ensure prices is 2D
        if prices.dim() == 1:
            prices = prices.unsqueeze(-1)
            
        # Calculate returns for mean reversion period
        mr_returns = change(prices, lag=self.mean_reversion_lookback)
        
        # Only keep momentum signals that align with mean reversion
        filtered_signals = signals.clone()
        filtered_signals[(signals > 0) & (mr_returns > 0)] = 0
        filtered_signals[(signals < 0) & (mr_returns < 0)] = 0
        
        return filtered_signals

    def _scale_positions(self, signals: torch.Tensor) -> torch.Tensor:
        """
        Scale positions over the holding period
        
        Args:
            - `signals` (torch.Tensor): Trading signals to scale
            
        Returns:
            - torch.Tensor: Scaled position weights
        """
        if not self.scale_positions:
            return signals
            
        weights = torch.zeros_like(signals)
        
        # For each day, build position over holding period
        for i in range(self.holding_period):
            shifted = F.pad(signals[:-i] if i > 0 else signals, (0, 0, i, 0))
            weights = weights + shifted
        
        # Normalize weights
        weights = weights / self.holding_period
        return weights

    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """
        Generate position weights based on momentum signals.
        
        Args:
            - `x` (torch.Tensor): Price tensor of shape (time,) or (time, assets)
            
        Returns:
            - torch.Tensor: Portfolio weights tensor between -1 and 1
        """
        # Store original shape
        original_dim = x.dim()
        
        # Ensure input is 2D
        if original_dim == 1:
            x = x.unsqueeze(-1)
            
        # Calculate returns
        returns = self._calculate_returns(x)
        
        # Generate signals
        signals = self._generate_signals(returns)
        
        # Apply mean reversion filter if enabled
        signals = self._apply_mean_reversion_filter(signals, x)
        
        # Scale positions if enabled
        weights = self._scale_positions(signals)
        
        # Return to original shape if input was 1D
        if original_dim == 1:
            weights = weights.squeeze(-1)
            
        return weights


class SimpleTimeSeriesMomentumModel(BaseAllocator):
    def __init__(self, lookback: int, hedge_market: bool = False):
        """
        Simple implementation of time series momentum strategy.
        
        Args:
            - `lookback` (int): Number of days to look back for return calculation
            - `hedge_market` (bool): Whether to neutralize market factor by subtracting average return
        """
        self.lookback = lookback
        self.hedge_market = hedge_market
        super().__init__(num_observations=lookback)
    
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """
        Generate position weights based on momentum signals.
        
        Args:
            - `x` (torch.Tensor): Price tensor of shape (time, assets)
            
        Returns:
            - torch.Tensor: Portfolio weights tensor
        """
        returns = change(x, lag=self.lookback)
        if self.hedge_market:
            market_return = returns.mean(dim=1)
            returns = returns - market_return.unsqueeze(1)
        return to_weights_matrix(returns)


class PriceChannelMomentumModel(BaseAllocator):
    """
    Price Channel Momentum strategy implementation.
    
    Strategy logic:
    1. Calculates price channel using either close prices only or high/low prices
    2. Takes long position (+1) when price breaks above upper channel
    3. Takes short position (-1) when price breaks below lower channel
    4. Holds position until price reaches channel midpoint
    """
    
    def __init__(
        self,
        lookback: int = 14,
        bounds: float = 0.8,
        use_high_low: bool = False
    ):
        """
        Args:
            - `lookback` (int): Period for calculating channel bounds
            - `bounds` (float): Width of the channel as percentage (0.0-1.0) from midpoint
            - `use_high_low` (bool): Whether to use high/low prices instead of just close
        """
        self.lookback = lookback
        self.bounds = bounds
        self.use_high_low = use_high_low
        super().__init__(num_observations=lookback)
        
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """
        Generate position weights based on price channel breakouts.
        
        Args:
            - `x` (torch.Tensor): Price tensor of shape (time, assets) if use_high_low=False
               or (time, assets, 3) with high/low/close if use_high_low=True
            
        Returns:
            - torch.Tensor: Portfolio weights tensor between -1 and 1
        """
        if self.use_high_low:
            # Handle 3D input (high, low, close)
            high = x[..., 0]
            low = x[..., 1]
            close = x[..., 2]
        else:
            # Use same prices for high/low/close
            high = close = x
            low = x  # Use same tensor for low to avoid None
        
        # Calculate channel bounds directly with tensors
        upper, lower = price_channel(
            high=high,
            low=low,  # Now always tensor instead of None
            lookback=self.lookback,
            bounds=self.bounds
        )
        
        # Calculate midpoint and generate signals
        midpoint = (upper + lower) / 2
        positions = multivariate_barrier(
            x=close,
            upper_bound=upper,
            lower_bound=lower,
            neutral_level=midpoint
        )
        
        # Normalize positions and maintain tensor device
        return positions / x.shape[1]

class VolatilityTargetingWrapper(BaseAllocator):
    """
    A wrapper class that applies volatility targeting to any allocator model.
    
    Volatility targeting aims to maintain a constant level of portfolio volatility by
    adjusting position sizes. This improves Sharpe ratio, reduces turnover, and manages risk.
    
    Source: https://www.algos.org/p/breaking-down-momentum-strategies
    
    Args:
        - `base_model` (BaseAllocator): The underlying allocation model to wrap
        - `volatility_period` (int): Lookback period for volatility calculation
        - `target_volatility` (float): Annual volatility target (e.g., 0.15 for 15%)
        - `max_leverage` (float): Maximum allowed leverage
        - `penalize_high_vol` (bool): Whether to additionally penalize high volatility assets
        - `T` (int): Number of days in a year (default: 365)
    """
    
    def __init__(
        self,
        base_model: BaseAllocator,
        volatility_period: int = 63,  # ~3 months of trading days
        target_volatility: float = 0.15,  # 15% annual volatility target
        max_leverage: float = 2.0,
        penalize_high_vol: bool = True,
        T: int = 365
    ):
        self.base_model = base_model
        self.volatility_period = volatility_period
        # Convert annual target to daily using specified T
        self.target_volatility = target_volatility / (T ** 0.5)  
        self.max_leverage = max_leverage
        self.penalize_high_vol = penalize_high_vol
        
        # Need at least as much data as the base model + volatility period
        required_observations = max(
            base_model.num_observations, 
            volatility_period
        )
        super().__init__(num_observations=required_observations)
    
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """
        Apply volatility targeting to the base model's weights
        
        Args:
            - `x` (torch.Tensor): Input tensor (can be prices or other data)
            
        Returns:
            Volatility-targeted weights
        """
        # Get base model weights
        base_weights = self.base_model(x)
        
        # Get price data for volatility calculations
        price_data = get_price_data(self.base_model, x)
        
        # Calculate asset volatilities
        asset_vols = calculate_asset_volatility(price_data, self.volatility_period)
        
        # Optionally penalize high volatility assets
        if self.penalize_high_vol:
            # Inverse volatility weighting
            vol_adjustment = 1.0 / (asset_vols + EPSILON)
            # Normalize
            vol_adjustment = vol_adjustment / vol_adjustment.sum()
            
            # Adjust base weights by volatility
            # Multiply by sign to preserve direction
            adjusted_weights = base_weights * torch.sign(base_weights) * vol_adjustment
            adjusted_weights = to_weights_matrix(adjusted_weights)
        else:
            adjusted_weights = base_weights
            
        # Calculate portfolio volatility based on recent returns
        if price_data.shape[0] > self.volatility_period + 1:
            # Calculate returns
            returns = change(price_data, lag=1)
            recent_returns = returns[-self.volatility_period:]
            
            # Use the most recent weights for volatility calculation
            last_weights = adjusted_weights
            if adjusted_weights.dim() > 1:
                last_weights = adjusted_weights[-1]
                
            portfolio_vol = calculate_portfolio_volatility(
                last_weights, 
                recent_returns
            )
            
            # Calculate scaling factor to achieve target volatility
            if portfolio_vol > 0:
                scaling_factor = self.target_volatility / portfolio_vol
                # Apply max leverage constraint
                scaling_factor = min(scaling_factor, self.max_leverage)
                
                # Apply scaling
                final_weights = adjusted_weights * scaling_factor
            else:
                final_weights = adjusted_weights
        else:
            final_weights = adjusted_weights
            
        return final_weights


class InformationDiscretenessFilter(BaseAllocator):
    """
    Filter assets based on Information Discreteness to focus on sustained price moves.
    
    Information Discreteness (ID) measures how evenly distributed price changes are.
    High ID means returns are concentrated in a few large moves.
    Low ID means returns are more evenly distributed.
    
    For momentum strategies, we prefer lower ID assets as they represent more sustained trends.
    
    Source: https://www.algos.org/p/breaking-down-momentum-strategies
    
    Parameters:
        base_model: The underlying allocation model
        lookback_period: Period for calculating ID
        id_threshold: Maximum allowed ID value (higher = more permissive)
        min_assets: Minimum number of assets to keep after filtering
    """
    
    ...  # TODO: implement


class RegimeSwitchingModel(BaseAllocator):
    """
    Switches between momentum and mean-reversion strategies based on market regime.
    
    Regime detection uses a simple method: if the market is above its moving average,
    it's considered a bull market (positive autocorrelation), suitable for momentum.
    If it's below, it's a bear market (negative autocorrelation), suitable for mean-reversion.
    
    Source: https://www.algos.org/p/breaking-down-momentum-strategies
    
    Parameters:
        momentum_model: Model to use in bull markets
        mean_reversion_model: Model to use in bear markets
        regime_period: Period for detecting market regime
        benchmark_index: Which asset to use as market benchmark (default: first asset)
        inverse_signals: Whether to invert signals in opposite regime rather than switching models
    """
    
    ...  # TODO: implement

class BetaFilteredMomentumModel(BaseAllocator):
    """
    Momentum strategy that filters out high-beta assets.
    
    From the article: "It is good practice to filter out the top decile assets by beta. 
    This is a rule that holds true for many other alphas. Even when ignoring the additional 
    turnover high beta assets tend to introduce, they still underperform on average and 
    increase the portfolio risk unnecessarily."
    
    Source: https://www.algos.org/p/breaking-down-momentum-strategies
    
    Parameters:
        lookback_period: Period for momentum calculation
        beta_measurement_period: Period for beta calculation
        beta_filter_threshold: Percentile to use as cutoff for high beta (e.g., 0.9 for top 10%)
        num_positions: Number of positions to hold
    """
    ...  # TODO: implement




import torch
import numpy as np
from typing import Callable

def _ensure_tensor(*args: (torch.Tensor| None)) -> list[torch.Tensor]: ...

def mid_price(bid_prices: torch.Tensor,
             ask_prices: torch.Tensor) -> torch.Tensor: ...

def spread(bid_prices: torch.Tensor,
          ask_prices: torch.Tensor,
          mid_price: torch.Tensor | None = None,
          relative: bool = False) -> torch.Tensor: ...

def imbalance_lob(bid_volumes: torch.Tensor,
                     ask_volumes: torch.Tensor) -> torch.Tensor: ...

def weighted_mid_price(bid_prices: torch.Tensor,
                      ask_prices: torch.Tensor,
                      imbalance_lob: torch.Tensor) -> torch.Tensor: ...

def weighted_average(values: torch.Tensor,
                    weights: torch.Tensor) -> torch.Tensor: ...

def vamp(bid_prices: torch.Tensor,
         ask_prices: torch.Tensor,
         bid_volumes: torch.Tensor,
         ask_volumes: torch.Tensor,
         target_quote_volume: float | list[float] | torch.Tensor,
         mid_price: torch.Tensor | None = None,
         relative: bool = False) -> torch.Tensor: ...

def reduce_wap(values: torch.Tensor,
              weights: torch.Tensor) -> torch.Tensor: ...

def vwap_orderbook_levels(prices: torch.Tensor,
                         volumes: torch.Tensor,
                         mid_price: torch.Tensor | None = None,
                         relative: bool = False) -> torch.Tensor: ...

def _process_weighted_price(func: Callable,
                          price: torch.Tensor,
                          volume: torch.Tensor,
                          window: int,
                          mp: torch.Tensor) -> torch.Tensor: ...

def all_features(bid_prices: torch.Tensor,
                ask_prices: torch.Tensor,
                bid_volumes: torch.Tensor,
                ask_volumes: torch.Tensor,
                vamp_levels: float | list[float] = 1e6,
                quote_asset_volume: torch.Tensor | None = None,
                normalize_vamp: bool = True,
                mad_deviations: float = 15
                ) -> list[torch.Tensor]: ...

def handle_relative_price(func: Callable) -> Callable: ...

def convert_to_quote_volumes(prices: torch.Tensor, volumes: torch.Tensor) -> torch.Tensor: ...

def create_volume_mask(volumes: torch.Tensor, target: float) -> torch.Tensor: ...

def calculate_masked_vwap(prices: torch.Tensor, 
                         volumes: torch.Tensor, 
                         mask: torch.Tensor) -> torch.Tensor: ...



from functools import wraps
from threading import RLock
from enum import Enum
from contextlib import contextmanager
from functools import _make_key
from collections import OrderedDict


class CacheMode(Enum):
    """
    Enumerate for cache mode.
    - `READ_WRITE`: Default: read from and write to cache
    - `READ_ONLY`: Read from cache, but do not write new values
    - `DISABLED`: Fully bypass cache
    """
    READ_WRITE = "READ_WRITE"
    READ_ONLY = "READ_ONLY"
    DISABLED = "DISABLED"

_CACHE_MODE = CacheMode.READ_WRITE

@contextmanager
def cache_mode(mode: CacheMode | str = CacheMode.READ_WRITE):
    """
    Caching context manager.
    Modes:
        - `READ_WRITE`
        - `READ_ONLY`
        - `DISABLED`
    """
    global _CACHE_MODE
    prev_mode = _CACHE_MODE
    _CACHE_MODE = mode
    try:
        yield
    finally:
        _CACHE_MODE = prev_mode

def conditional_lru_cache(maxsize=128, typed=False):
    def decorator(func):        
        cache = {}
        lock = RLock()
        # Use OrderedDict for O(1) access and LRU tracking
        cache_order = OrderedDict()

        @wraps(func)
        def wrapper(*args, **kwargs):
            key = _make_key(args, kwargs, typed)
            global _CACHE_MODE

            if _CACHE_MODE == CacheMode.DISABLED:
                return func(*args, **kwargs)

            with lock:
                if key in cache:
                    # Move to end (most recently used) in O(1)
                    cache_order.move_to_end(key)
                    return cache[key]

                try:
                    result = func(*args, **kwargs)
                except Exception:
                    # Don't cache failed function calls
                    raise

                if _CACHE_MODE == CacheMode.READ_WRITE:
                    cache[key] = result
                    cache_order[key] = None  # Add to end
                    
                    # Evict least recently used if cache is full
                    if len(cache_order) > maxsize:
                        oldest_key = next(iter(cache_order))
                        del cache[oldest_key]
                        del cache_order[oldest_key]

                return result

        def cache_clear():
            with lock:
                cache.clear()
                cache_order.clear()

        wrapper.cache_clear = cache_clear
        return wrapper

    return decorator



import torch

from datetime import datetime
import numpy as np

from MLTT.utils import apply_columnwise, CACHE_SIZE
from MLTT.cache import conditional_lru_cache


@conditional_lru_cache(CACHE_SIZE)
def sma(close_log: torch.Tensor, period: int) -> torch.Tensor:
    """
    Computes Simple Moving Average using cumulative sum approach.
    Handles edge cases where window size exceeds available data.
    
    Args:
        - `close_log` (torch.Tensor): Input tensor of shape (T, N) where T is time dimension
        - `period` (int): SMA window size (must be positive)
        
    Returns:
        Tensor: SMA with same shape as input, dtype float32
    """
    if period <= 0:
        raise ValueError(f"Period must be positive, got {period}")
    if close_log.ndim != 2:
        raise ValueError(f"Input must be 2D tensor (T, N), got shape {close_log.shape}")
    
    T, N = close_log.shape
    if T == 0:
        return close_log.clone()
    
    cumsum = torch.cumsum(close_log, dim=0)
    
    if period > T:
        window = torch.arange(1, T+1, device=close_log.device, dtype=torch.float32).view(-1, 1)
        return (cumsum / window).to(close_log.dtype)
    
    # Calculate moving sum using shifted cumulative sums
    shifted = torch.zeros_like(cumsum)
    shifted[period:] = cumsum[:-period]
    moving_sum = cumsum - shifted
    
    # Calculate dynamic window sizes (clipped at period)
    window = torch.arange(1, T+1, device=close_log.device, dtype=torch.float32).view(-1, 1)
    window = torch.minimum(window, torch.tensor(period, dtype=torch.float32, device=close_log.device))
    
    return (moving_sum / window).to(close_log.dtype)


@conditional_lru_cache(CACHE_SIZE)
def ema(close_log: torch.Tensor, period: int) -> torch.Tensor:
    """
    Computes Exponential Moving Average (EMA) of a time series using vectorized PyTorch ops.

    Args:
        - `close_log` (torch.Tensor): Time series (T, N), where T — time, N — series.
        - `period` (int): Smoothing period.

    Returns:
        torch.Tensor: EMA of same shape as input.
    """
    alpha = 2 / (period + 1)
    close_log = close_log.float()
    ema = torch.zeros_like(close_log)
    ema[0] = close_log[0]  # init first value to be same

    # recursive EMA computation, vectorized over columns
    for t in range(1, close_log.shape[0]):
        ema[t] = alpha * close_log[t] + (1 - alpha) * ema[t - 1]
    
    return ema

@conditional_lru_cache(CACHE_SIZE)
def stdev(series: torch.Tensor, period: int) -> torch.Tensor:
    """
    Calculates rolling standard deviation of a time series.

    Args:
        - `series` (torch.Tensor): (T, N) tensor where T — time, N — different series (columns).
        - `period` (int): Rolling window size.

    Returns:
        torch.Tensor: (T, N) rolling standard deviation. First few rows will be NaN or zero due to padding.
    """
    T = series.shape[0]
    if period > T:
        raise ValueError("Period can't be greater than time dimension")

    # pad the front so that output shape stays (T, N)
    pad = period - 1
    padded = torch.nn.functional.pad(series.T.unsqueeze(1), (pad, 0), mode='replicate')  # (N, 1, T + pad)

    # unfold to get rolling windows
    unfolded = padded.unfold(dimension=2, size=period, step=1)  # (N, 1, T, period)
    
    # calc std over last dim (window)
    std = unfolded.std(dim=-1)  # (N, 1, T)
    
    return std.squeeze(1).T  # -> (T, N)


@conditional_lru_cache(CACHE_SIZE)
def _kalman_muth_ewma_volatility(returns, 
                                mu: float = 0.0, 
                                lambda_: float = 0.05,
                                tau_psi: float = 0.02, 
                                tau_eta: float = 0.1,
                                init_state_variance: float = 1.0,
                                ):
    """
    Kalman filter for estimating volatility using a state-space model
    with mean-reversion in the latent log-variance process.

    Muth's EWMA model (1960).
    Source: The elements of quantitative investing (2024), Giuseppe A. Paleologo
    
    Args:
        - `returns` (torch.Tensor): Array of returns.
        - `mu` (float): Long-term mean of variance (log-space).
        - `lambda_` (float): Mean-reversion strength.
        - `tau_psi` (float): Std of process noise.
        - `tau_eta` (float): Std of observation noise.
        - `init_state_variance` (float): Initial state variance (log-space) for the Kalman filter.

    Returns:
        - volatility_estimates (np.ndarray): Estimated volatilities (std).
        - log_variance_estimates (np.ndarray): Estimated log-variances.
    """
    n = len(returns)
    
    # Initialize state variables
    x_pred = mu  # initial estimate of log variance
    P_pred = init_state_variance  # initial state variance

    # Storage
    x_filt = torch.zeros(n)
    P_filt = torch.zeros(n)
    sigma_est = torch.zeros(n)

    # Precompute constants
    gamma = -1.2704  # E[log(χ²_1)] ≈ -1.27
    R = tau_eta ** 2  # Observation noise variance
    Q = tau_psi ** 2  # Process noise variance

    for t in range(n):
        # Observation: y_t = log(r_t^2) - gamma
        y_t = torch.log(returns[t]**2 + 1e-8) - gamma  # small value added to avoid log(0)

        # --- Prediction step ---
        x_pred = (1 - lambda_) * x_pred + lambda_ * mu
        P_pred = (1 - lambda_)**2 * P_pred + Q

        # --- Update step ---
        K = P_pred / (P_pred + R)  # Kalman gain
        x_upd = x_pred + K * (y_t - x_pred)
        P_upd = (1 - K) * P_pred

        # Save results
        x_filt[t] = x_upd
        P_filt[t] = P_upd
        sigma_est[t] = torch.exp(x_upd / 2)

        # Prepare for next step
        x_pred = x_upd
        P_pred = P_upd

    return sigma_est, x_filt

@conditional_lru_cache(CACHE_SIZE)
def kalman_muth_volatility(
        returns, 
        mu: float = 0.0, 
        lambda_: float = 0.05,
        tau_psi: float = 0.02, 
        tau_eta: float = 0.1,
        init_state_variance: float = 1.0,
    ):
    if returns.dim() == 1:
        return _kalman_muth_ewma_volatility(returns, mu, lambda_, tau_psi, tau_eta, init_state_variance)[0]
    return apply_columnwise(returns, lambda x: _kalman_muth_ewma_volatility(x, mu, lambda_, tau_psi, tau_eta, init_state_variance)[0])


@conditional_lru_cache(CACHE_SIZE)
def z_score(series: torch.Tensor, period: int, mean: float | None = None) -> torch.Tensor:
    if mean is None:
        return (series - sma(series, period)) / stdev(series, period)
    else:
        return (series - mean) / stdev(series, period)


@conditional_lru_cache(CACHE_SIZE)
def vwap(prices: torch.Tensor, volumes: torch.Tensor, period: int) -> torch.Tensor:
    """
    Computes Volume-Weighted Average Price with dynamic window handling.
    Implements both cumulative and rolling window approaches.
    
    Args:
        - `prices` (torch.Tensor): Tensor of shape (T,) or (T, N) containing price values
        - `volumes` (torch.Tensor): Tensor of shape (T,) or (T, N) with corresponding volumes
        - `period` (int): Lookback window size for VWAP calculation
        
    Returns:
        Tensor: VWAP values with same shape as input
    """
    if prices.shape != volumes.shape:
        raise ValueError(f"Prices and volumes must have same shape. Got {prices.shape} vs {volumes.shape}")
    
    if period <= 0:
        raise ValueError(f"Period must be positive, got {period}")
    
    # Calculate price-volume product and cumulative sums
    pv = prices * volumes
    cum_pv = torch.cumsum(pv, dim=0)
    cum_vol = torch.cumsum(volumes, dim=0)
    
    # Handle full-period windows using rolling difference
    if len(prices) > period:
        shifted_pv = torch.zeros_like(cum_pv)
        shifted_pv[period:] = cum_pv[:-period]
        
        shifted_vol = torch.zeros_like(cum_vol)
        shifted_vol[period:] = cum_vol[:-period]
        
        window_pv = cum_pv - shifted_pv
        window_vol = cum_vol - shifted_vol
    else:
        window_pv = cum_pv
        window_vol = cum_vol
    
    # Add epsilon to prevent division by zero
    epsilon = torch.finfo(prices.dtype).eps
    vwap_values = window_pv / (window_vol + epsilon)
    
    # Maintain original dtype
    return vwap_values.to(prices.dtype)


@conditional_lru_cache(CACHE_SIZE)
def rsi(close_log: torch.Tensor, period: int) -> torch.Tensor:
    """
    Calculates the Relative Strength Index (RSI) of a time series.

    Args:
        - `close_log` (torch.Tensor): Tensor of shape (T, N) containing price values
        - `period` (int): Lookback window size for RSI calculation
        
    Returns:
        Tensor: RSI values with same shape as input
    """
    T, N = close_log.shape

    delta = close_log[1:] - close_log[:-1]  # (T-1, N)

    gain = torch.clamp(delta, min=0)
    loss = -torch.clamp(delta, max=0)

    avg_gain = torch.zeros((T - 1, N), device=close_log.device)
    avg_loss = torch.zeros((T - 1, N), device=close_log.device)

    avg_gain[period - 1] = gain[:period].mean(dim=0)
    avg_loss[period - 1] = loss[:period].mean(dim=0)

    for t in range(period, T - 1):
        avg_gain[t] = (avg_gain[t - 1] * (period - 1) + gain[t]) / period
        avg_loss[t] = (avg_loss[t - 1] * (period - 1) + loss[t]) / period

    rs = avg_gain / (avg_loss + 1e-8)
    rsi = 100 - 100 / (1 + rs)

    # Padding to match the input length
    rsi_full = torch.zeros((T, N), device=close_log.device)
    rsi_full[period:] = rsi[period - 1:]  # first period values are default 0

    return rsi_full



class GlobalTimeFeatures:
    def __init__(self, timestamp: torch.Tensor):
        # Convert timestamp tensor to datetime objects
        self.datetime = np.array([
            datetime.fromtimestamp(ts.item())
            for ts in timestamp / 1000
        ])
    
    def day_of_week(self) -> torch.Tensor:
        # Returns tensor with values 0 (Monday) through 6 (Sunday)
        return torch.tensor([
            dt.weekday() 
            for dt in self.datetime
        ], dtype=torch.long).unsqueeze(-1)
    
    def day_of_month(self) -> torch.Tensor:
        # Returns tensor with values 1-31
        return torch.tensor([
            dt.day 
            for dt in self.datetime
        ], dtype=torch.long).unsqueeze(-1)
    
    def month_of_year(self) -> torch.Tensor:
        # Returns tensor with values 1-12
        return torch.tensor([
            dt.month 
            for dt in self.datetime
        ], dtype=torch.long).unsqueeze(-1)
    
    def hour_of_day(self) -> torch.Tensor:
        # Returns tensor with values 0-23
        return torch.tensor([
            dt.hour 
            for dt in self.datetime
        ], dtype=torch.long).unsqueeze(-1)


"""
Data structures for alpha factors.
"""
from dataclasses import dataclass

from torch import Tensor


@dataclass
class AlphaData:
    """Container for alpha factor data."""
    
    values: Tensor  # Shape: (n_observations, n_features)
    feature_names: list[str]
    
    def __post_init__(self):
        assert len(self.feature_names) == self.values.shape[1], \
            "Feature names length must match number of features: " \
            f"{len(self.feature_names)} != {self.values.shape[1]}"
    
    @property
    def n_features(self) -> int:
        return len(self.feature_names)
    
    def get_feature_idx(self, name: str) -> int:
        """Get index of feature by name."""
        try:
            return self.feature_names.index(name)
        except ValueError:
            raise ValueError(f"Feature {name} not found in data")



#TODO: Rewrite this models using torch
from dataclasses import dataclass

import numpy as np
from tqdm import auto
import torch

from .kalman import RegressionModel
from .mean_reversion import ZScoreBarriersModel
from MLTT.allocation.allocators import BaseAllocator
from MLTT.utils import to_weights_matrix


@dataclass
class RegressionIndices:
    endog_idx: int
    exog_indices: list[int]


def _get_exog_indices(index_map: np.ndarray) -> list[RegressionIndices]:
    """
    Generates a mapping of regression variables (endog) and factors (exog).

    Args:
        - `index_map` (numpy.ndarray): A matrix where each row represents a regression model.
            The first column is the index of the variable, and the rest of
            the columns are indicators (0 or 1) of whether the endogenous
            variable at the corresponding index uses the exogenous variable
            at the same index as a predictor.

    Returns:
        - `list[RegressionIndices]`: A list of tuples, where each
            tuple contains the index of the endogenous variable and a
            list of the indices of the exogenous variables.
    """
    indices = []
    for i in range(index_map.shape[0]):
        exog_row = index_map[i]
        endog_idx = int(index_map[i, 0])
        exog_row_indices = np.flatnonzero(exog_row[1:])
        indices.append(RegressionIndices(endog_idx, exog_row_indices.tolist()))

    return indices


def _get_endog_indices(index_map: np.ndarray | list[RegressionIndices]) -> list[int]:
    if isinstance(index_map, np.ndarray):
        index_map = _get_exog_indices(index_map)

    endog_indices = []
    for indices in index_map:
        endog_indices.append(indices.endog_idx)

    return endog_indices


def update_step(x: np.ndarray, models: list[RegressionModel], exog_indices) -> list[np.ndarray]:
    """
    Updates the regression coefficients for each regression model using the given data point.

    Args:
        - `x` (numpy.ndarray): The data point to use for updating the regression coefficients
        - `models` (list[RegressionModel]): A list of regression models to update
        - `exog_indices` (list[RegressionIndices]): A list of tuples,
            where each tuple contains the index of the endogenous
            variable and a list of the indices of the exogenous variables

    Returns:
        - list[numpy.ndarray]: A list of the updated regression coefficients
            for each regression model
    """
    states: list[np.ndarray] = []

    for i, (endog_idx, exog_index) in enumerate(exog_indices):
        states.append(
            models[i].update(
                x[exog_index],
                x[endog_idx]
            )
        )
    return states


def all_combinations(n_exog) -> np.ndarray: # returns indices matrix
    """
    Generates a matrix of all possible combinations of endogenous and exogenous variables.

    Args:
        - `n_exog` (int): The number of exogenous variables

    Returns:
        - numpy.ndarray: A matrix where each row
            represents a regression model.
            The first column is the index of the endogenous
            variable, and the rest of the columns are the
            indices of the exogenous variables
    """
    exog_matrix = np.ones((n_exog, n_exog))
    exog_matrix -= np.eye(n_exog)

    endog_idx = np.arange(n_exog).reshape(-1, 1)

    return np.hstack([endog_idx, exog_matrix])


def regress_predictions(prices_matrix, models, exog_indices, use_tqdm=True, save_states=True):
    """
    Generates and saves the predictions and states matrices.

    Args:
        - `prices_matrix` (numpy.ndarray): A matrix of prices for
            each asset in the portfolio
        - `models` (list[RegressionModel]): A list of regression models
        - `exog_indices` (list[RegressionIndices]): A list of
            tuples, where each tuple contains the index of the endogenous
            variable and a list of the indices of the exogenous variables
        - `use_tqdm` (bool): Whether to use tqdm for progress bar
        - `save_states` (bool): Whether to save the historical states

    Returns:
        - tuple[numpy.ndarray, list[numpy.ndarray]]: A tuple containing
            the predictions matrix and a list of states matrices
    """
    predictions = np.zeros((prices_matrix.shape[0], len(exog_indices)))
    states: list[np.ndarray] = [
        np.zeros((prices_matrix.shape[0], len(indices.exog_indices)))
        for indices in exog_indices
    ]

    iterator = enumerate(models)
    if use_tqdm:
        iterator = auto.tqdm(iterator, total=len(models))

    # using GPU-accelerated functions
    for i, model in iterator:
        exog = prices_matrix[:, exog_indices[i].exog_indices]
        endog = prices_matrix[:, exog_indices[i].endog_idx]
        predictions[:, i] = model.run(exog, endog)
        if save_states:
            states[i] = model.states


    return predictions, states


def predictions_by_coefficients(prices_matrix, coefficients, indices):
    """
    Calculate predictions using coefficient matrices.
    
    Args:
        - `prices_matrix` (numpy.ndarray): Matrix of prices
        - `coefficients` (numpy.ndarray): Coefficient values
        - `indices` (numpy.ndarray): Indices to use from prices matrix
        
    Returns:
        - numpy.ndarray: Predicted values
    """
    return np.sum(prices_matrix[:, indices] * coefficients, axis=1)


def many_hot_decode(dim_raw_matrix: int, states: np.ndarray, indices: np.ndarray) -> np.ndarray:
    """
    Transforms a matrix of coefficients from the size of the
    indices to the size of the raw matrix of observations.

    Args:
        - `dim_raw_matrix` (int): The size of the raw matrix of observations
        - `states` (numpy.ndarray): A matrix of coefficients of shape (n_observations, n_exog)
        - `indices` (numpy.ndarray): A matrix of indices of shape (n_exog,)

    Returns:
        - numpy.ndarray: A matrix of coefficients of shape (n_observations, dim_raw_matrix)
    """
    raw = np.zeros((states.shape[0], dim_raw_matrix))
    raw[:, indices] = states
    return raw


class VectorRegressionModel:
    def __init__(self,
                 exog_indices: list[RegressionIndices] | np.ndarray,
                 models: list[RegressionModel],
                 use_tqdm=True) -> None:
        """
        Initialize a vector regression model with multiple individual regression models.
        
        Args:
            - `exog_indices` (list[RegressionIndices] | np.ndarray): Indices of exogenous variables
            - `models` (list[RegressionModel]): Regression models for each endogenous variable
            - `use_tqdm` (bool): Whether to use tqdm progress bars
        """
        if not isinstance(exog_indices, list):
            exog_indices = _get_exog_indices(exog_indices)

        self.indices = exog_indices
        self.models = models
        self.use_tqdm = use_tqdm

    def update(self, x: np.ndarray) -> list[np.ndarray]:
        """
        Update all regression models with new data.
        
        Args:
            - `x` (numpy.ndarray): New data point for updating models
            
        Returns:
            - list[numpy.ndarray]: List of updated model states
        """
        return update_step(x, self.models, self.indices)

    def run(self, x: np.ndarray) -> tuple[np.ndarray, list[np.ndarray]]:
        """
        Runs the vector regression model on the given data.

        Args:
            - `x` (numpy.ndarray): An array of shape (n_observations, n_variables)
                containing the data to be used for prediction

        Returns:
            - tuple[numpy.ndarray, list[numpy.ndarray]]: A tuple containing the predictions
            and the states of each regression model

            - `predictions` (numpy.ndarray): An array of
              shape (n_observations, n_exog) containing
              the predictions for each regression model
            - `states` (list[numpy.ndarray]): A list of
              arrays of shape (n_observations, n_exog) containing
              the states of each regression model
        """
        predictions, states = regress_predictions(
            x,
            self.models,
            self.indices,
            use_tqdm=self.use_tqdm
        )
        return predictions, states

    def predict(self, x: np.ndarray) -> np.ndarray:
        """
        Make predictions with all regression models for a single observation.
        
        Args:
            - `x` (numpy.ndarray): Input data for prediction
            
        Returns:
            - numpy.ndarray: Predictions for each model
        """
        predictions = np.zeros(len(self.models))

        for i, model in enumerate(self.models):
            prediction = model.predict(x)
            if prediction.ndim != 1 or prediction.shape[0] != 1:
                raise ValueError(
                    f"Predictions must be 1D, got {prediction.shape}-dim. Error in model {i}"
                )

            predictions[i] = prediction.item()

        return predictions

    def fit(self, x: np.ndarray):
        """
        Fit all regression models on historical data.
        
        Args:
            - `x` (numpy.ndarray): Historical data for training models
        """
        for model, indices in auto.tqdm(zip(self.models, self.indices), total=len(self.models)):
            endog = x[:, indices.endog_idx]
            exog = x[:, indices.exog_indices]

            model.fit(exog, endog)

class OneLegIndexSignalModel(BaseAllocator):
    def __init__(self,
                 z_score_model: ZScoreBarriersModel,
                 vector_regression: VectorRegressionModel) -> None:
        """
        A model that combines regression and z-score for signal generation.
        
        Args:
            - `z_score_model` (ZScoreBarriersModel): Model for generating signals based on z-scores
            - `vector_regression` (VectorRegressionModel): Model for predicting fair prices through regression
        """
        self._z_model = z_score_model
        self._regression = vector_regression
        self._endog_idx = _get_endog_indices(vector_regression.indices)
        self.num_observations = 1

    def fit(self, x: np.ndarray) -> None:
        """
        Fits all regression models on the data.

        Args:
            - `x` (np.ndarray): Matrix with log-prices for regression
                of shape `(n_observations, n_tradable)`
        """
        self._regression.fit(x)

    def predict(self, x: np.ndarray) -> np.ndarray:
        """
        Generates base weights using log-prices.

        Args:
            - `x` (np.ndarray): Matrix with log-prices
                Shape: `(n_observations, n_tradable)`

        Returns:
            - np.ndarray: Weights in base assets
                Shape: `(n_observations, n_tradable)`
        """
        prediction, _ = self._regression.run(x)
        spreads = x - prediction
        signals = self._z_model.predict(spreads)
        return signals


class TorchRefitModel(BaseAllocator):
    def __init__(self, 
                 model: torch.nn.Module,
                 lookback: int = 5,
                 n_refit: int = 100,
                 n_refit_data: int = 500,
                 device: str = "cpu",
                 lr: float = 0.01,
                 refit_epochs: int = 3):
        """
        A model that periodically refits a PyTorch neural network model.
        
        Args:
            - `model` (torch.nn.Module): PyTorch model to use for predictions
            - `lookback` (int): Window size for input features
            - `n_refit` (int): How often to refit the model (in steps)
            - `n_refit_data` (int): How many historical data points to use when refitting
            - `device` (str): Device to run model on ("cpu" or "cuda")
            - `lr` (float): Learning rate for optimizer
            - `refit_epochs` (int): Number of epochs to train when refitting
        """
        super().__init__()
        # The model is moved to the specified device
        self.model = model.to(device)
        # Other parameters are initialized
        self.lookback = lookback
        self.n_refit = n_refit
        self.n_refit_data = n_refit_data
        self.device = device
        self.lr = lr
        self.refit_epochs = refit_epochs
        
        # The optimizer and loss function are initialized
        self.optimizer = torch.optim.SGD(self.model.parameters(), lr=lr)
        self.criterion = torch.nn.MSELoss()
        
        # Other variables are initialized
        self.step_counter = 0
        self.history = torch.tensor([], device=device)
        self.last_processed_idx = 0
        self.initial_padding = None  # For storing initial padding

    def _prepare_window(self, data: torch.Tensor) -> torch.Tensor:
        """
        Creates sliding windows directly in tensors.
        
        Args:
            - `data` (torch.Tensor): Input data tensor
            
        Returns:
            - torch.Tensor: Tensor with sliding windows
        """
        return data.unfold(0, self.lookback, 1).flatten(start_dim=1)

    def _refit_model(self, data: torch.Tensor):
        """
        Fully tensor-based training pipeline.
        
        Args:
            - `data` (torch.Tensor): Historical data for training
        """
        self.model.train()
        
        # Data preparation
        X = self._prepare_window(data[:-1])
        y = data[self.lookback-1:-1]
        
        # Training loop
        for _ in range(self.refit_epochs):
            self.optimizer.zero_grad()
            outputs = self.model(X)
            loss = self.criterion(outputs, y)
            loss.backward()
            self.optimizer.step()

    def predict(self, x: np.ndarray) -> torch.Tensor:
        """
        Predict using model with periodical refitting.
        
        Args:
            - `x` (np.ndarray): Input data tensor
            
        Returns:
            - torch.Tensor: Predicted weights
        """
        if x.shape[0] < self.lookback:
            raise ValueError(f"Need at least {self.lookback} observations, got {x.shape[0]}")
            
        x_tensor = torch.as_tensor(x, dtype=torch.float32, device=self.device)
        
        # Padding initialization on the first call
        if self.history.numel() == 0:
            self.initial_padding = torch.zeros(
                (self.lookback-1, x.shape[1]), 
                device=self.device
            )
            self.history = torch.cat([self.initial_padding, x_tensor], dim=0)
            self.last_processed_idx = self.lookback - 1
        else:
            self.history = torch.cat([self.history, x_tensor], dim=0)
        
        start_idx = self.last_processed_idx + 1
        end_idx = self.history.size(0)
        
        predictions = []
        for i in range(start_idx, end_idx):
            if (i - self.lookback) % self.n_refit == 0 and i >= self.n_refit_data:
                train_data = self.history[i-self.n_refit_data:i]
                self._refit_model(train_data)
            
            window = self.history[i-self.lookback:i].view(-1)
            
            with torch.no_grad():
                self.model.eval()
                pred = self.model(window)
                current_price = self.history[i]
                predictions.append(pred - current_price)
        
        self.last_processed_idx = end_idx - 1
        
        # Assemble the full result with padding
        full_predictions = torch.cat([
            self.initial_padding if self.initial_padding is not None else torch.tensor([]),
            torch.stack(predictions) if predictions else torch.tensor([], device=self.device)
        ])
        
        # Trim to the size of the input data
        result = full_predictions[-x.shape[0]:]
        
        # Reset padding after the first call
        self.initial_padding = None
        
        return to_weights_matrix(result)


# This is a simple linear model for fair price prediction
class SimpleLinearFairPrice(torch.nn.Module):
    def __init__(self, n_assets: int, lookback: int):
        """
        Initialize a simple linear model for fair price prediction.
        
        Args:
            - `n_assets` (int): Number of assets in the portfolio
            - `lookback` (int): Number of historical observations to use for prediction
        """
        super().__init__()
        self.linear = torch.nn.Linear(n_assets * lookback, n_assets)
        
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass through the model.
        
        Args:
            - `x` (torch.Tensor): Input tensor of shape (batch_size, n_assets * lookback)
            
        Returns:
            - torch.Tensor: Predicted fair prices
        """
        # Add squeeze to remove batch dimension
        return self.linear(x).squeeze(0)



from MLTT.data_loading.dumper import dump_data, load_csv_files_from_directory, load_klines_from_file
from MLTT.data_loading.formatter import format_files, sort_dates
from MLTT.data_loading.interface import (load_and_format, 
                                         save_symbols, 
                                         load_symbols, 
                                         only_format, 
                                         ccxt_ohlcv, 
                                         check_klines_timeline,
                                         check_linearity)
from MLTT.data_loading.selection import make_klines_array
from MLTT.data_loading.defillama import DefiLlamaAdapter



"""
Base classes for alpha factor representation.
"""
from abc import ABC, abstractmethod
from dataclasses import dataclass, fields

from torch import Tensor


class Operation(ABC):
    """Base class for all operations in alpha factor trees."""
    
    @abstractmethod
    def forward(self, x: Tensor) -> Tensor:
        """Apply operation to input tensor."""
        pass
    
    @abstractmethod
    def to_string(self) -> str:
        """Convert operation to string representation."""
        pass
    
    @classmethod
    @abstractmethod
    def from_string(cls, s: str) -> 'Operation':
        """Create operation from string representation."""
        pass


@dataclass
class AlphaFactor:
    """Represents a complete alpha factor as a tree of operations."""
    
    root: Operation
    feature_names: list[str]
    
    def compute(self, data: Tensor) -> Tensor:
        """Compute alpha factor values for given input data.
        
        Args:
            - `data` (Tensor): Input tensor of shape (n_observations, n_features)
            
        Returns:
            - `Tensor`: Tensor of shape (n_observations,) with alpha factor values
        """
        assert data.shape[1] == len(self.feature_names), \
            f"Expected {len(self.feature_names)} features, got {data.shape[1]}"
        return self.root.forward(data)
    
    def to_string(self) -> str:
        """Convert alpha factor to string representation."""
        return f"{self.root.to_string()}|{','.join(self.feature_names)}"
    
    @classmethod
    def from_string(cls, s: str) -> 'AlphaFactor':
        """Create alpha factor from string representation."""
        from .operations import Operation, InputFeature
        
        op_str, features_str = s.split('|')
        feature_names = features_str.split(',')
        
        # Create alpha factor
        alpha = cls(
            root=Operation.from_string(op_str),
            feature_names=feature_names
        )
        
        # Fix feature indices in InputFeature operations
        def fix_indices(op: Operation) -> None:
            if isinstance(op, InputFeature):
                op.feature_idx = feature_names.index(op.feature_name)
            for field in fields(op):
                value = getattr(op, field)
                if isinstance(value, Operation):
                    fix_indices(value)
                elif isinstance(value, (list, tuple)):
                    for item in value:
                        if isinstance(item, Operation):
                            fix_indices(item)
        
        fix_indices(alpha.root)
        return alpha 


from MLTT.allocation.backtesting import BTResult, backtest_model, backtest
from MLTT.allocation.backend import allocation_quote_profit
from MLTT.allocation.allocators import (CapitalAllocator,
                         BaseAllocator,
                         BlendingModel)



from MLTT.allocation import BaseAllocator
from MLTT.utils import to_weights_matrix
from MLTT.technical.features import z_score
from MLTT.utils import apply_columnwise, change, multivariate_barrier, EPSILON
from MLTT.models.momentum import SimpleTimeSeriesMomentumModel

import torch
from MLTT.utils import calculate_betas


class ZScoreBarriersModel(BaseAllocator):
    def __init__(self,
                 period: int,
                 lower_band: float,
                 upper_band: float,
                 all_in_position: bool = False,
                 mean: float = None) -> None:
        self.period = period
        self.upper_band = upper_band
        self.lower_band = lower_band
        self.all_in_position = all_in_position
        self.mean = mean

        super().__init__(num_observations=period if not mean else 1)

    def predict(self, x: torch.Tensor) -> torch.Tensor:
        z = z_score(x, self.period, self.mean)
        signals = multivariate_barrier(z, self.upper_band, self.lower_band)

        if self.all_in_position:
            weights = to_weights_matrix(signals)
        else:
            weights = signals / x.shape[1]

        return weights


class ZScorePortionModel(BaseAllocator):
    def __init__(self,
                 period: int,
                 mean: float = None) -> None:
        self.period = period
        self.mean = mean

        super().__init__(num_observations=period if not mean else 1)

    def predict(self, x: torch.Tensor) -> torch.Tensor:
        z = z_score(x, self.period, self.mean)

        return to_weights_matrix(-z).nan_to_num(0)


class CrossSectionalMeanReversionModel(BaseAllocator):
    """
    Cross-sectional mean reversion model.
    Source: Ernest P. Chan, "Algorithmic Trading"
    """

    def __init__(self, num_observations: int = 1, hedge_market: bool = True) -> None:
        self._momentum = SimpleTimeSeriesMomentumModel(
            lookback=num_observations, 
            hedge_market=hedge_market
        )
    
    @property
    def min_observations(self) -> int:
        return self._momentum.min_observations

    def predict(self, x: torch.Tensor) -> torch.Tensor:
        return -self._momentum.predict(x)

class EnhancedCrossSectionalMRModel(BaseAllocator):
    """
    Enhanced Cross-Sectional Mean Reversion Model with Volatility Filter.
    Based on the article by Teddy Koker.
    URL: https://teddykoker.com/2019/05/improving-cross-sectional-mean-reversion-strategy-in-python/
    """

    def __init__(self, 
                 num_positions: int, 
                 lag: int, 
                 volatility_period: int) -> None:
        self.num_positions = num_positions
        self.lag = lag
        self.volatility_period = volatility_period
        super().__init__(num_observations=max(lag, volatility_period))

    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """
        Generate equal-weighted positions based on mean reversion.
        
        Args:
            x: Price matrix of shape (time, assets)
            
        Returns:
            weights: Portfolio weights with shape (time, assets)
        """
        # Calculate returns using change
        returns = change(x, lag=self.lag)
        
        # Calculate mean returns across assets
        mean_returns = torch.mean(returns, dim=1, keepdim=True)
        
        # Calculate weights based on the formula
        deviations = returns - mean_returns
        weights = -deviations
        
        # Initialize result array with zeros
        final_weights = torch.zeros_like(weights)
        
        # Process each time step separately
        for t in range(weights.shape[0]):
            # Calculate volatility for current timestep
            if t < self.volatility_period:
                vol = torch.std(x[:t+1], dim=0) if t > 0 else torch.ones(x.shape[1])
            else:
                vol = torch.std(x[t-self.volatility_period+1:t+1], dim=0)
            
            # Get current weights
            curr_weights = weights[t]
            
            # Find top positions by absolute weight
            _, top_by_weight = torch.sort(torch.abs(curr_weights), descending=True)
            top_by_weight = top_by_weight[:self.num_positions]
            
            # Find top positions by lowest volatility
            _, top_by_vol = torch.sort(vol)
            top_by_vol = top_by_vol[:self.num_positions]
            
            # Find intersection of indices
            top_by_weight_set = set(top_by_weight.tolist())
            top_by_vol_set = set(top_by_vol.tolist())
            selected = torch.tensor(list(top_by_weight_set.intersection(top_by_vol_set)), dtype=torch.long)
            
            if len(selected) > 0:
                # Get selected weights
                selected_weights = curr_weights[selected]
                
                # Normalize selected weights
                norm_factor = torch.sum(torch.abs(selected_weights))
                if norm_factor > 0:
                    selected_weights = selected_weights / norm_factor
                    
                # Assign normalized weights
                for i, idx in enumerate(selected):
                    final_weights[t, idx] = selected_weights[i]
        
        return final_weights


class BetaAdjustedMeanReversionModel(BaseAllocator):
    """
    Beta-adjusted mean reversion model.
    Quant. Arb: https://www.algos.org/p/breaking-down-momentum-strategies
    """
    def __init__(self, num_observations: int = 1, beta_measurement_period: int = 200) -> None:
        super().__init__(num_observations=max(num_observations, beta_measurement_period))
        self.beta_measurement_period = beta_measurement_period
        self.num_observations = num_observations

    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """
        Args:
            - x: tensor of shape (num_observations, num_assets+1)
            the last column is the benchmark
        """
        # Separate the assets and benchmark
        assets = x[:, :-1]
        benchmark = x[:, -1:]
        
        # Calculate asset returns
        asset_returns = change(assets, lag=self.num_observations)
        
        # Calculate benchmark returns
        bench_returns = change(benchmark, lag=self.num_observations)
        
        # Calculate betas using the extracted function
        betas = calculate_betas(asset_returns, bench_returns, self.beta_measurement_period)
        
        # Apply beta adjustment to returns
        beta_adjusted_returns = torch.zeros_like(asset_returns)
        for i in range(assets.shape[1]):
            # Divide returns by beta to get beta-adjusted returns
            # Add epsilon to avoid division by zero
            beta_adjusted_returns[:, i] = asset_returns[:, i] / (betas[i] + EPSILON)
        
        # Calculate mean of beta-adjusted returns across assets
        mean_returns = beta_adjusted_returns.mean(dim=1, keepdim=True)
        
        # Calculate weights based on deviations from mean (mean reversion)
        deviations = beta_adjusted_returns - mean_returns
        
        # Return the negative deviations as weights (buy underperformers, sell outperformers)
        return to_weights_matrix(-deviations)



"""
Data -> technical features -> ML allocator (fitting to autograd sharpe or other) -> weights
-> backtesting -> gradient
"""
from torch import Tensor, abs, cat, sum, ones, any, stack, zeros, roll, zeros_like

from warnings import warn

from MLTT.utils import EPSILON, change, CACHE_SIZE
from MLTT.cache import conditional_lru_cache


@conditional_lru_cache(maxsize=CACHE_SIZE)
def add_neutral_weight(x):
    """
    Adds a neutral weight to each asset.

    Args:
        - `x` (Tensor): A tensor of shape `(batch_size, n_tradable)`,
            where `batch_size` is the number of samples and
            `n_tradable` is the number of tradable assets.

    Returns:
        Tensor: A tensor of shape `(batch_size, n_tradable+1)`,
            where `batch_size` is the number of samples,
            `n_tradable` is the number of tradable assets and
            the last column contains neutral weights,
            in which the sum of each row is strictly equal to 1.
    """
    sums = sum(abs(x), dim=1)
    neutral = 1 - sums

    return cat((x, neutral.unsqueeze(1)), dim=1)

@conditional_lru_cache(maxsize=CACHE_SIZE)
def add_neutral_price(x):
    """
    Adds a neutral price to each asset.

    Args:
        - `x` (Tensor): A tensor of shape `(batch_size, n_tradable)`,
            where `batch_size` is the number of samples and
            `n_tradable` is the number of tradable assets.

    Returns:
        Tensor: A tensor of shape `(batch_size, n_tradable+1)`,
            where `batch_size` is the number of samples,
            `n_tradable` is the number of tradable assets and
            the last column contains neutral prices.
    """
    return cat((x, ones(x.shape[0], 1, device=x.device)), dim=1)

@conditional_lru_cache(maxsize=CACHE_SIZE)
def _validate_weights(weights):
    """
    Calls `warn` if some weights are invalid (>1 in sum).

    Args:
        - `weights` (Tensor): Weights matrix of shape: `(batch_size, n_tradable)`.
    """
    s = sum(abs(weights), dim=1) - EPSILON
    if any(s > 1):
        warn("invalid weights in allocation")

@conditional_lru_cache(maxsize=CACHE_SIZE)
def _process_backtest_inputs(
        weights: Tensor,
        prices: Tensor,
        prices_change: Tensor | None = None,
        check_weights: bool = True
    ) -> tuple[Tensor, Tensor, Tensor]:
    """
    Common input processing for allocation functions.

    Args:
        - `weights` (Tensor): Weights matrix
        - `prices` (Tensor): Prices matrix  
        - `prices_change` (Tensor): Logarithmic returns of prices
        - `check_weights` (bool): Whether to validate weights

    Returns:
        tuple[Tensor, Tensor, Tensor]: Processed weights, prices and price changes
    """
    weights, prices = reshape_predictions(weights, prices)
    
    if check_weights:
        _validate_weights(weights)

    prices = add_neutral_price(prices)
    
    if prices_change is None:
        prices_change = change(prices)
    else:
        prices_change = stack((
            prices_change, 
            zeros(prices_change.shape[0], 1, 
                  device=prices_change.device)
        ), dim=1)
        
    weights = add_neutral_weight(weights)
    
    return weights, prices, prices_change


@conditional_lru_cache(maxsize=CACHE_SIZE)
def allocation_quote_profit(weights: Tensor,
                            prices: Tensor,
                            prices_change: Tensor | None = None, 
                            commission: float = 0.0,
                            check_weights: bool = True):
    """
    Function calculates profit from quote weights without any adjustment in time.
    Note: This function provides approximate results and should only be used for rough 
    strategy evaluation.

    No market impact or slippage is considered.

    Args:
        - `weights` (Tensor): Weights matrix of shape: `(batch_size, n_tradable)`.
            Sum of absolute values in each row is `<=1` (negative values used for short positions).
            Needs to be already shifted to the next time step to avoid leakage in backtesting.
        - `prices` (Tensor): Prices matrix of shape: `(batch_size, n_tradable)`.
        - `prices_change` (Tensor): Logarithmic returns of prices.
            Shape: `(batch_size, n_tradable)`.
        - `commission` (float): Commission rate in range `[0, 1]`.
        - `check_weights` (bool): Whether to validate input weights.
            If the weights are invalid, function will issue a warning.

    Returns:
        tuple[Tensor, Tensor]: A tuple containing:
            - Gross returns before commissions
            - Commission costs as positive relative values
    """
    weights, prices, prices_change = _process_backtest_inputs(
        weights,
        prices, 
        prices_change, 
        check_weights)

    change = sum(weights * prices_change, dim=1)
    
    commission_abs = zeros_like(change)
    if commission:
        rolled_quote_weights = roll(weights, shifts=1, dims=0)
        quote_weights_diff = weights - rolled_quote_weights
        commission_abs = commission * sum(abs(quote_weights_diff), dim=1)

    return change, commission_abs


@conditional_lru_cache(maxsize=CACHE_SIZE)
def reshape_predictions(weights: Tensor, prices: Tensor) -> tuple[Tensor, Tensor]:
    """
    Reshapes one-dim array to two-dim array if needed.
    
    Args:
        - `weights` (Tensor): model output weights.
            Shape: `(batch_size, n_tradable)`.
        - `prices` (Tensor): prices matrix.
            Shape: `(batch_size, n_tradable)`.

    Returns:
        Tensor: weights matrix of shape: `(batch_size, n_tradable)`.
        Tensor: prices matrix of shape: `(batch_size, n_tradable)`.
    """
    if len(weights.shape) == 1:
        weights = weights.reshape(-1, 1)
    if len(prices.shape) == 1:
        prices = prices.reshape(-1, 1)
    return weights, prices



from abc import ABC, abstractmethod

import numpy as np
from torch import Tensor, ones, zeros_like
import scipy.stats as ss
from scipy.optimize import minimize
from MLTT.utils import ensure_tensor


class RegressionModel(ABC):
    @abstractmethod
    def predict(self, x: Tensor) -> Tensor:
        """
        Predict output for input data.
        
        Args:
            - `x` (Tensor): Input data
            
        Returns:
            - Tensor: Predicted values
        """
        ...
        
    @abstractmethod
    def fit(self, x: Tensor, y: Tensor):
        """
        Fit model to training data.
        
        Args:
            - `x` (Tensor): Input features
            - `y` (Tensor): Target values
        """
        ...

    @abstractmethod
    def update(self, x: Tensor, y: Tensor):
        """
        Update model with new data point.
        
        Args:
            - `x` (Tensor): Input features
            - `y` (Tensor): Target value
        """
        ...

    @abstractmethod
    def run(self, x: Tensor, y: Tensor) -> Tensor:
        """
        Run model on data sequence.
        
        Args:
            - `x` (Tensor): Input features sequence
            - `y` (Tensor): Target values sequence
            
        Returns:
            - Tensor: Predictions
        """
        ...

    @abstractmethod
    def calibrate(self, x: Tensor, y: Tensor):
        """
        Calibrate model parameters.
        
        Args:
            - `x` (Tensor): Input features
            - `y` (Tensor): Target values
        """
        ...

class AdaptiveLinearRegression(RegressionModel):
    """
    This class implements adaptive linear regression with a Kalman filter.

    Attributes:
        - `num_features` (int): Number of features in the data
        - `Q` (Tensor): State transition noise covariance matrix (diagonal)
        - `R` (Tensor): Measurement noise covariance matrix (diagonal)
        - `F` (float): Forgetting factor (0 <= F <= 1)
        - `P` (Tensor): Initial covariance matrix
        - `state` (Tensor): Initial regression coefficients
        - `l2_reg` (float): L2 regularization parameter (default: 0)
    """

    def __init__(self, Q, R, F=0.999, P=None, state=None, l2_reg=0, save_states=True):
        """
        Initializes the AdaptiveLinearRegression object.

        Args:
            - `Q` (Tensor): State transition noise covariance matrix (diagonal)
            - `R` (Tensor): Measurement noise covariance matrix
            - `F` (float): Forgetting factor (0 <= F <= 1).
                Brian Lai and Dennis S. Bernstein (arXiv:2404.10914v1)
            - `P` (Tensor | None): Initial covariance matrix. Defaults to None
            - `state` (Tensor | None): Initial regression coefficients. Defaults to None
            - `l2_reg` (float): L2 regularization parameter. Defaults to 0
            - `save_states` (bool): Whether to save the historical state of the Kalman filter. Defaults to True
        """
        self.num_features = Q.shape[0]
        self.Q = Q
        self.R = R
        self.F = F
        if P is None:
            P = Q/2 # Initial covariance (identity matrix)
            # by the paper Q-P need to be positive definite
        if state is None:
            state = ones(self.num_features)/self.num_features  # Initial coefficients
        self.P = P
        self.state = state
        self.l2_reg = l2_reg

        self.pdf = ss.norm.pdf
        self.log_likelihood = 0

        self.save_states = save_states
        self.states = None

    def predict(self, x: Tensor) -> Tensor:
        """
        Predicts the target value for a given input data point.

        Args:
            - `x` (Tensor): Input data point of shape (num_features,)

        Returns:
            - Tensor: Predicted target value
        """
        return x @ self.state

    def update(self, x: Tensor, y: Tensor) -> Tensor:
        """
        Updates the regression coefficients based on a new data point.

        Args:
            x (Tensor): Input data point (shape: (num_features,)).
            y (float): Target value for the data point.

        Returns:
            Tensor: Updated regression coefficients.
        """
        # using numba function
        self.P, self.state = update(
            self.P,
            self.state,
            x,
            y,
            self.Q,
            self.R,
            self.F,
            l2_reg=self.l2_reg,
        )

        return self.state

    def run(self, x: Tensor, y: Tensor) -> Tensor:
        """
        Runs the Kalman filter on the input data.

        Args:
            x (Tensor): Input data (shape: (num_observations, num_features)).
            y (Tensor): Target values (shape: (num_observations,)).

        Returns:
            Tensor: Array of predicted target values.
                Shape: (num_observations,)
        """
        if x.shape[0] != y.shape[0]:
            raise ValueError(
                "x and y must have the same number of observations,"
                f"but got x.shape={x.shape} and y.shape={y.shape}"
            )

        self.log_likelihood = 0

        # using numba function
        states, self.P, predictions = adaptive_linear_regression(
            self.Q,
            self.R,
            self.F,
            self.P,
            self.state,
            x,
            y,
            l2_reg=self.l2_reg,
            save_states=self.save_states,
        )
        if self.save_states:
            self.state = states[-1]
            self.states = states

        return predictions

    def fit(self, x: Tensor, y: Tensor):
        """
        Fits the model to the training data using ordinary least squares.

        Args:
            x (Tensor): Training data (shape: (num_observations, num_features)).
            y (Tensor): Target values (shape: (num_observations,)).
        """
        # Handle target values with a single column (if needed)
        if y.ndim == 2 and y.shape[1] == 1:
            y = y.flatten()  # Reshape to 1D array

        self.state = np.linalg.lstsq(x, y, rcond=None)[0]
        self.run(x, y)

    def calibrate(self, x: Tensor, y: Tensor): # TODO: debug. This method does not converge
        """
        Calibrates the model using Maximum Likelihood Estimation (MLE).

        Args:
            x (Tensor): Training data (shape: (num_observations, num_features)).
            y (Tensor): Target values (shape: (num_observations,)).
        """
        def neg_log_likelihood(params):
            Q = np.diag(params[:self.num_features])
            R = np.diag(params[self.num_features:])
            self.Q = Q
            self.R = R
            self.fit(x, y)
            self.run(x, y)
            return -self.log_likelihood

        initial_params = np.concatenate((np.diag(self.Q), np.diag(self.R)))

        res = minimize(
            neg_log_likelihood,
            initial_params,
            method="L-BFGS-B",
            bounds=[(1e-10, None) for _ in range(2 * self.num_features)],
            tol=1e-6,
        )
        self.Q = np.diag(res.x[:self.num_features])
        self.R = np.diag(res.x[self.num_features:])
        if res.success:
            self.log_likelihood = -res.fun
        else:
            raise RuntimeError(f"Optimization failed: {res.message}")

def update(P: Tensor, state: Tensor, x: Tensor, y: Tensor, Q: Tensor, R: Tensor, F: float, l2_reg: float):
    # Convert inputs to tensors if they are numpy arrays
    P, state, x, y, Q, R = ensure_tensor(P, state, x, y, Q, R)

    # Predict target value for the input data point 
    y_hat = x @ state

    # Calculate the innovation (prediction error)
    v = y - y_hat

    # Update the covariance matrix with the forgetting factor
    P = F * P + Q

    # Calculate the Kalman gain
    S = x.T @ P @ x + R
    K = P @ x.T @ S.inverse()  # Kalman gain

    # Update the regression coefficients
    state += K.T * v - l2_reg * state

    y_hat = x @ state

    return P, state, y_hat

def adaptive_linear_regression(
        Q: Tensor, 
        R: Tensor,
        F: float, 
        P: Tensor, 
        state: Tensor, 
        x: Tensor, 
        y: Tensor, 
        l2_reg: float = 0, 
        save_states: bool = False
    ) -> tuple[Tensor, Tensor, Tensor]:
    """
    Implements adaptive linear regression with a Kalman filter.

    Args:
        Q (torch.Tensor or numpy.ndarray): State transition noise covariance matrix (diagonal).
        R (torch.Tensor or numpy.ndarray): Measurement noise covariance matrix (diagonal).
        F (float): Forgetting factor (0 <= F <= 1).
        P (Tensor): Initial covariance matrix.
        state (Tensor): Initial regression coefficients.
        x (Tensor): Input data (shape: (num_observations, num_features)).
        y (Tensor): Target values (shape: (num_observations,)).
        l2_reg (float, optional): L2 regularization parameter. Defaults to 0.
        save_states (bool, optional): Whether to save the historical state of the Kalman filter. Defaults to False.

    Returns:
        - torch.Tensor: Array of updated regression coefficients.
            Shape: (num_observations, num_features)
        - torch.Tensor: Updated covariance matrix.
        - torch.Tensor: Array of predicted target values.
    """
    # Convert inputs to tensors if they are numpy arrays
    Q, R, P, state, x, y = ensure_tensor(Q, R, P, state, x, y)

    num_observations = x.shape[0]
    states = zeros_like(x)
    predictions = zeros_like(y)

    for i in range(num_observations):
        Pi = P.clone()
        statei = state.clone()
        P, state, predictions[i] = update(Pi, statei, x[i].clone(), y[i], Q, R, F, l2_reg=l2_reg)
        if save_states:
            states[i] = state

    return states, P, predictions



# TODO: SWITCH TO TIMESCALEDB
import torch
import polars as pl
import pandas as pd
import pickle

from dataclasses import dataclass

@dataclass
class Orderbook:
    bid_prices: torch.Tensor
    bid_sizes: torch.Tensor
    ask_prices: torch.Tensor
    ask_sizes: torch.Tensor
    received_time: torch.Tensor

    @classmethod
    def from_parquet(cls, path, levels=10):
        orderbook = pl.read_parquet(path)
        return orderbook_to_vectors(orderbook, levels)

    def to_pickle(self, path, verbose=True):
        with open(path, "wb") as f:
            pickle.dump(self, f)
        if verbose:
            print(f"Orderbook saved successfully to {path}")
    
    @classmethod
    def from_pickle(cls, path, verbose=True):
        with open(path, "rb") as f:
            orderbook = pickle.load(f)
        if verbose:
            print(f"Orderbook loaded successfully from {path}")
        return orderbook


def parse_tensor_column(column, strip=None):
    """
    Parse a Polars Series into tensors, truncating to `strip` length if provided.
    Fills missing values with appropriate defaults.

    Args:
        column (pl.Series): Polars column with tensor-like data in string format.
        strip (int, optional): Maximum elements in each tensor. Defaults to None.

    Returns:yi
        torch.Tensor: Stacked tensor of all rows.
    """
    arrays = column.to_numpy().astype(str)  # Convert to numpy strings
    tensor_data = []
    
    for arr in arrays:
        values = arr.split(",")[:strip]
        # Convert values to float, replacing invalid with 0
        float_values = []
        for x in values:
            try:
                float_values.append(float(x))
            except ValueError:
                float_values.append(0.0)
                
        # Pad with zeros if not enough values
        if strip and len(float_values) < strip:
            if "price" in column.name:
                # For prices, repeat last valid price
                last_valid = float_values[-1] if float_values else 0.0
                float_values.extend([last_valid] * (strip - len(float_values)))
            else:
                # For sizes, pad with zeros
                float_values.extend([0.0] * (strip - len(float_values)))
                
        tensor_data.append(torch.tensor(float_values, dtype=torch.float32))
        
    return torch.stack(tensor_data)

def orderbook_to_vectors(orderbook: pl.DataFrame, levels: int = 10):
    """
    Convert order book data into a dictionary of tensors.
    Handles missing or invalid data by filling with appropriate values.

    Args:
        orderbook (pl.DataFrame): Polars DataFrame containing:
            - 'bid_prices', 'bid_sizes', 'ask_prices', 'ask_sizes'.
        levels (int): Maximum number of price/size levels.

    Returns:
        Dict[str, torch.Tensor]: Dictionary containing parsed tensors.
    """
    # Convert timestamps to nanoseconds
    received_time_ns = pd.to_datetime(orderbook["received_time"]).astype('int64')

    return Orderbook(
        bid_prices=parse_tensor_column(orderbook["bid_prices"], strip=levels),
        bid_sizes=parse_tensor_column(orderbook["bid_sizes"], strip=levels),
        ask_prices=parse_tensor_column(orderbook["ask_prices"], strip=levels),
        ask_sizes=parse_tensor_column(orderbook["ask_sizes"], strip=levels),
        received_time=torch.as_tensor(received_time_ns.to_numpy(), dtype=torch.int64)
    )



from warnings import warn
from functools import wraps

import torch

from MLTT.utils import mad_normalize, ensure_tensor, CACHE_SIZE
from MLTT.cache import conditional_lru_cache


BPS_NORM = 1e4  # 1 basis point is 1/10_000 of the price

def handle_relative_price(func):
    """
    Decorator to handle relative price calculations
    
    Args:
        func: Function that returns a tensor of shape (batch_size,) or (batch_size, n_levels)
        
    Returns:
        Wrapped function that handles relative price calculations
    """
    @wraps(func)
    def wrapper(*args, mid_price=None, relative=False, **kwargs):
        result = func(*args, **kwargs)
        if relative and mid_price is not None:
            if mid_price.dim() == 2:
                mid_price = mid_price.squeeze(-1)
            if result.dim() > mid_price.dim():
                mid_price = mid_price.unsqueeze(-1)
            result = result / mid_price - 1
        return result
    return wrapper

@conditional_lru_cache(maxsize=CACHE_SIZE)
def weighted_average(values, weights):
    """
    Calculate simple weighted average along last dimension
    """
    values, weights = ensure_tensor(values, weights)
    return (values * weights).sum(dim=-1) / (weights.sum(dim=-1) + 1e-10)

@conditional_lru_cache(maxsize=CACHE_SIZE)
def convert_to_quote_volumes(prices, volumes):
    """
    Convert base volumes to quote volumes
    """
    return volumes * prices

@conditional_lru_cache(maxsize=CACHE_SIZE)
def create_volume_mask(volumes, target):
    """
    Create mask for volumes up to target
    """
    cum_vol = torch.cumsum(volumes, dim=-1)
    return cum_vol <= target

@conditional_lru_cache(maxsize=CACHE_SIZE)
def calculate_masked_vwap(prices, volumes, mask):
    """
    Calculate VWAP using masked prices and volumes
    """
    masked_prices = prices * mask
    masked_volumes = volumes * mask
    return weighted_average(masked_prices, masked_volumes)

@conditional_lru_cache(maxsize=CACHE_SIZE)
@handle_relative_price
def vamp(bid_prices, ask_prices, bid_volumes, ask_volumes, target_quote_volume):
    """
    Calculate Volume-Adjusted Mid Price (VAMP) based on target quote volume threshold(s)
    
    Args:
        bid_prices: tensor of shape (batch_size, depth) 
        ask_prices: tensor of shape (batch_size, depth)
        bid_volumes: bid volumes tensor of shape (batch_size, depth)
        ask_volumes: ask volumes tensor of shape (batch_size, depth)
        target_quote_volume: target volume threshold(s) in quote currency (e.g. USD)
            Can be a single number or list/array/tensor of thresholds
        
    Returns:
        VAMP tensor of shape:
            - (batch_size,) if target_quote_volume is a single number
            - (batch_size, n_levels) if target_quote_volume is a sequence

    Citation:
         Martin, Payton and Line Jr., William and Feng, Yuxin and Yang, Yunfan and Zheng, 
            Sharon and Qi, Susan and Zhu, Beiming, Mind the Gaps: Short-Term Crypto Price 
            Prediction (December 15, 2022). Available at SSRN:
            https://ssrn.com/abstract=4351947 or http://dx.doi.org/10.2139/ssrn.4351947 
    """
    bid_prices, ask_prices, bid_volumes, ask_volumes, target_quote_volume = ensure_tensor(
        bid_prices, ask_prices, bid_volumes, ask_volumes, target_quote_volume)
    if target_quote_volume.dim() == 0:
        target_quote_volume = target_quote_volume.unsqueeze(0)
    
    quote_bid_volumes = convert_to_quote_volumes(bid_prices, bid_volumes)
    quote_ask_volumes = convert_to_quote_volumes(ask_prices, ask_volumes)
    
    vamps = []
    for target_vol in target_quote_volume:
        bid_mask = create_volume_mask(quote_bid_volumes, target_vol)
        ask_mask = create_volume_mask(quote_ask_volumes, target_vol)
        
        vwap_bid = calculate_masked_vwap(bid_prices, quote_bid_volumes, bid_mask)
        vwap_ask = calculate_masked_vwap(ask_prices, quote_ask_volumes, ask_mask)
        
        vamp_ = (vwap_bid + vwap_ask) / 2
        vamps.append(vamp_)
    
    vamps = torch.stack(vamps, dim=-1)
    return vamps.squeeze() if len(target_quote_volume) == 1 else vamps

@conditional_lru_cache(maxsize=CACHE_SIZE)
def mid_price(bid_prices, ask_prices):
    """
    Calculate mid price between bid and ask prices
    """
    bid, ask = ensure_tensor(bid_prices, ask_prices)
    return (bid + ask) / 2

@conditional_lru_cache(maxsize=CACHE_SIZE)
def spread(bid_prices, ask_prices, mid_price=None, relative=False):
    bid, ask, mp = ensure_tensor(bid_prices, ask_prices, mid_price)
    if relative:
        return (ask - bid) / mp
    return ask - bid

@conditional_lru_cache(maxsize=CACHE_SIZE)
def imbalance_lob(bid_volumes, ask_volumes):
    """
    Calculate limit order book imbalance for each level.

    Imbalance = (bid_volumes) / (total_volumes) 
    In range [0, 1] where 0 means absolute balance and 1 means buying pressure.
    
    Args:
        bid_volumes: bid volumes tensor of shape (batch_size, depth)
        ask_volumes: ask volumes tensor of shape (batch_size, depth)
        
    Returns:
        Tensor of shape (batch_size, depth) where each element [i,j] represents
        limit order book imbalance calculated using volumes from levels 0 to j for batch element i
    """
    bid, ask = ensure_tensor(bid_volumes, ask_volumes)
    cumsum_bid = torch.cumsum(bid, dim=-1)
    cumsum_ask = torch.cumsum(ask, dim=-1)
    
    denominator = cumsum_bid + cumsum_ask
    denominator = torch.where(denominator == 0, torch.ones_like(denominator), denominator)
    
    return cumsum_bid / denominator

@conditional_lru_cache(maxsize=CACHE_SIZE)
def weighted_mid_price(bid_prices, ask_prices, imbalance_lob):
    """
    Calculate weighted mid price from bid and ask prices and imbalance.

    Args:
        bid_prices: tensor of shape (batch_size, depth)
        ask_prices: tensor of shape (batch_size, depth)
        imbalance_lob: tensor of shape (batch_size, depth)

    Returns:
        Tensor of shape (batch_size, depth)
    """
    bid_prices, ask_prices, imbalance_lob = ensure_tensor(bid_prices, ask_prices, imbalance_lob)
    return bid_prices * (1 - imbalance_lob) + ask_prices * imbalance_lob

@conditional_lru_cache(maxsize=CACHE_SIZE)
def reduce_wap(values, weights):
    """
    Calculate weighted average price (common logic for VWAP and TWAP)
    
    For each position n in output, calculates weighted average of first n elements 
    from values using first n elements from weights as weights.
    """
    if values.dim() != weights.dim():
        warn(f"Values and weights have different dimensions. values: {values.shape}, weights: {values.shape}")
    
    values_cumsum = torch.cumsum(values * weights, dim=-1)
    weights_cumsum = torch.cumsum(weights, dim=-1)
    weights_cumsum = torch.where(weights_cumsum == 0, torch.ones_like(weights_cumsum), weights_cumsum)
    
    return values_cumsum / weights_cumsum

@conditional_lru_cache(maxsize=CACHE_SIZE)
def vwap_orderbook_levels(prices, volumes, mid_price=None, relative=False):
    prices, volumes, mid_price = ensure_tensor(prices, volumes, mid_price)
    vwap_ = reduce_wap(prices, volumes)
    if relative:
        vwap_ /= mid_price
        vwap_ -= 1
    return vwap_

@conditional_lru_cache(maxsize=CACHE_SIZE)
def all_features(bid_prices, 
                 ask_prices, 
                 bid_volumes, 
                 ask_volumes, 
                 vamp_levels=1e6,
                 normalize_vamp: bool = True,
                 mad_deviations: float = 15,
                 relative: bool = True):
    """
    Calculate all orderbook features
    
    Args:
        bid_prices: tensor of shape (batch_size, depth)
        ask_prices: tensor of shape (batch_size, depth) 
        bid_volumes: tensor of shape (batch_size, depth)
        ask_volumes: tensor of shape (batch_size, depth)
        vamp_levels: target quote volume for VAMP calculation (list[float] or float)
            Can be a single number or list/array/tensor of quote volumes. 
            Default is 1e6 (1M USD e.g.)
        normalize_vamp: whether to apply MAD normalization to VAMP (default: True)
        mad_deviations: number of MADs to use as threshold for outliers (default: 15)
        relative: whether to return relative values (default: True)
    Returns:
        List of tensors containing orderbook features:
        - mid price (shape: batch_size, depth)
        - imbalance (shape: batch_size, depth)
        - weighted mid price (shape: batch_size, depth)
        - spread (shape: batch_size, depth)
        - VAMP (shape: batch_size, n_levels)
        - VWAP (shape: batch_size, depth)
    """
    bid_prices, ask_prices, bid_volumes, ask_volumes = ensure_tensor(
        bid_prices, ask_prices, bid_volumes, ask_volumes)
    
    # Calculate mid price (shape: batch_size, depth)
    mp = mid_price(bid_prices, ask_prices)
    
    # Calculate spread in bps
    spread_ = spread(bid_prices, ask_prices, mp[:, 0].unsqueeze(-1), relative=relative)
    if relative:
        spread_ *= BPS_NORM
    
    # Calculate limit order book imbalance
    imb_lob = imbalance_lob(bid_volumes, ask_volumes)
    
    # Calculate weighted mid price
    wmp = weighted_mid_price(bid_prices, ask_prices, imb_lob)
    
    # Calculate VAMP for multiple levels
    vamp_ = vamp(bid_prices, ask_prices, bid_volumes, ask_volumes,
                mid_price=mp[:, 0], target_quote_volume=vamp_levels, relative=relative)
    if relative:
        vamp_ *= BPS_NORM
    
    # Normalize VAMP if requested
    if normalize_vamp:
        vamp_ = mad_normalize(vamp_, deviations=mad_deviations, axis=1)
    
    # Calculate VWAP for orderbook levels
    vwap = vwap_orderbook_levels(bid_prices, bid_volumes, mid_price=mp[:, 0].unsqueeze(-1), relative=relative)
    if relative:
        vwap *= BPS_NORM

    # Return list of features
    return [
        mp,          # [batch_size, depth]
        imb_lob,     # [batch_size, depth]
        wmp,         # [batch_size, depth]
        spread_,     # [batch_size, depth] 
        vamp_,       # [batch_size, n_levels]
        vwap,        # [batch_size, depth]
    ]
